/*
 * amscm.h
 *
 * This file is part of AM-Scheme.
 * AM-Scheme is coded by Atsushi Moriwaki
 *
 * This software is completely free to copy and distribute modifed version.
 * But I would appreciate it if you left my name on the code as the author.
 *
 */
 
#include "config.h"

typedef unsigned short uShort;

#define T_NULL		0
#define T_STRING	1
#define T_INTEGER	2
#define T_REAL		3
#define T_COMPLEX	4
#define T_CHAR		5
#define T_SYMBOL	6
#define T_SYNTAX	7
#define T_PRIM1		8
#define T_PRIM2		9
#define T_PAIR		10
#define T_VECTOR	11
#define T_CLOSURE	12
#define T_CONTINUATION	13
#define T_MACRO		14
#define T_INPUT		15
#define T_OUTPUT	16
#define T_BOX		17
typedef int	cell_type;


/* dummy cell structure */
struct dummy_cell {
    cell_type _type;
    struct dummy_cell *_new;
};
typedef struct dummy_cell *pointer;
#define type(p)          ((p)->_type)
#define new_alloc(p)     ((p)->_new)
#define isreallocated(p) (type(p) == T_NULL)
#define BOXSIZE          (long)(sizeof(struct dummy_cell))
#define isbox(p)         (type(p) == T_BOX)
#define boxcontext(p)    ((p)->_new)

/* cons cell structure */
struct _cons_cell {
    cell_type _type;
    pointer _car; 
    pointer _cdr;    
};
typedef struct _cons_cell *cons_cell;
#define CONSSIZE        (long)(sizeof(struct _cons_cell))
#define ispair(p)       (type(p) == T_PAIR)
#define car(p)          (((cons_cell)(p))->_car)
#define cdr(p)          (((cons_cell)(p))->_cdr)
#define isclosure(p)    (type(p) == T_CLOSURE)
#define ismacro(p)      (type(p) == T_MACRO)
#define closure_code(p) car(p)
#define closure_env(p)  cdr(p)

/* integer cell structure */
struct _int_cell {
    cell_type _type;
    char _sign;
    long _size;
    uShort _array[1];
};
typedef struct _int_cell *int_cell;
#define INTEGERSIZE(l) \
((long)sizeof(struct _int_cell)+(long)((l)-1)*sizeof(uShort))
#define INTEGERBASE  10000
#define isinteger(p)    (type(p) == T_INTEGER)
#define integersign(p)  (((int_cell)(p))->_sign)
#define integersize(p)  (((int_cell)(p))->_size)
#define integerarray(p) (((int_cell)(p))->_array)

/* real cell structure */
struct _real_cell {
    cell_type _type;
    double _rvalue;
};
typedef struct _real_cell *real_cell;
#define REALSIZE        (long)(sizeof(struct _real_cell))
#define isreal(p)       (type(p) == T_REAL)
#define rvalue(p)       (((real_cell)(p))->_rvalue)

/* char cell structure */
struct _char_cell {
    cell_type _type;
    uShort _cvalue;
};
typedef struct _char_cell *char_cell;
#define ischar(p)       (type(p) == T_CHAR)
#define cvalue(p)       (((char_cell)(p))->_cvalue)

/* string cell structure */
struct _str_cell {
    cell_type _type;
    long _len;
    char _svalue[1];
};
typedef struct _str_cell *str_cell;
# define STRINGSIZE(l)    \
((long)sizeof(struct _str_cell)+(long)(l)*sizeof(char))
#define isstring(p)     (type(p) == T_STRING)
#define strvalue(p)     (((str_cell)(p))->_svalue)
#define strlength(p)    (((str_cell)(p))->_len)

/* symbol cell structure */
struct _symbol_cell {
    cell_type _type;
    int _op;
    pointer _property;
    long _len;
    char _svalue[1];
};
typedef struct _symbol_cell *symbol_cell;
#define SYMBOLSIZE(l)    \
((long)sizeof(struct _symbol_cell)+(long)(l)*sizeof(char))
#define issymbol(p)     ((type(p) == T_SYMBOL)||(type(p) == T_SYNTAX))
#define symlen(p)       (((symbol_cell)(p))->_len)
#define symname(p)      (((symbol_cell)(p))->_svalue)
#define symprop(p)      (((symbol_cell)(p))->_property)
#define hasprop(p)      ((type(p) == T_SYMBOL)||(type(p) == T_SYNTAX))
#define issyntax(p)     (type(p) == T_SYNTAX)
#define syntaxlen(p)    (((symbol_cell)(p))->_len)
#define syntaxname(p)   (((symbol_cell)(p))->_svalue)
#define syntaxprop(p)   (((symbol_cell)(p))->_property)
#define syntaxnum(p)    (((symbol_cell)(p))->_op)

/* vector cell structure */
struct _vector_cell {
    cell_type _type;
    long _len;
    pointer _vec[1];
};
typedef struct _vector_cell *vector_cell;
#define VECTORSIZE(l)    \
((long)sizeof(struct _vector_cell)+(long)((l)-1)*sizeof(pointer))
#define isvector(p)     (type(p) == T_VECTOR)
#define vector_len(p)   (((vector_cell)(p))->_len)
#define vector_vec(p)   (((vector_cell)(p))->_vec)
#define iscontinuation(p) (type(p) == T_CONTINUATION)
#define cont_len(p)     vector_len(p)
#define cont_copy(p)    vector_vec(p)

/* primitive cell structure */
struct _primitive_cell {
    cell_type _type;
    union {
        int _op;
        pointer (*_proc)();
    } _prim_context;
    int _nargs;
#define	AT_LEAST_N(x)	(-(x))
#define	MUST_BE_N(x)	((x) + 1)
    char _name[1];
};
typedef struct _primitive_cell *primitive_cell;
#define PRIMITIVESIZE(l)    \
((long)sizeof(struct _primitive_cell)+(long)(l)*sizeof(char))
#define isprim1(p)      (type(p) == T_PRIM1)
#define isprim2(p)      (type(p) == T_PRIM2)
#define isprim(p)       (type(p) == T_PRIM1 || type(p) == T_PRIM2)
#define primname(p)     (((primitive_cell)(p))->_name)
#define primargs(p)	(((primitive_cell)(p))->_nargs)
#define primnum(p)      (((primitive_cell)(p))->_prim_context._op)
#define primproc(p)     (((primitive_cell)(p))->_prim_context._proc)

/* input port structure */
struct _input_cell {
    cell_type _type;
    int status;
    int index;
    long pos;
    long eol;
#ifndef THINK_C_DA
    FILE *fp;
#else
    int ref;
    int linestatus;
    char *other_buffer;
#endif
    char buffer[LINESIZE];
};
typedef struct _input_cell *input_cell;
#define INPUTSIZE       (long)(sizeof(struct _input_cell))
#define isinput(p)      (type(p) == T_INPUT)
#define inputstatus(p)  (((input_cell)(p))->status)
#define inputindex(p)   (((input_cell)(p))->index)
#define inputpos(p)     (((input_cell)(p))->pos)
#define inputeol(p)     (((input_cell)(p))->eol)
#define inputbuffer(p)  (((input_cell)(p))->buffer)
#ifndef THINK_C_DA
#define inputfile(p)    (((input_cell)(p))->fp)
#define clear_inputline(p) { inputpos(p) = inputeol(p) = 0; }
#else
#define inputfile(p)    (((input_cell)(p))->ref)
#define inputlinest(p)  (((input_cell)(p))->linestatus)
#define inputotherbuff(p) (((input_cell)(p))->other_buffer)
#define isactiveline(p)  (inputlinest(p) != 0)
#define EOF (-1)
#define clear_inputline(p) { inputpos(p) = inputeol(p) = inputlinest(p) = 0; }
#endif

/* output port structure */
struct _output_cell {
    cell_type _type;
    int status;
    int index;
#ifndef THINK_C_DA
    FILE *fp;
#else
    int ref;
    int xpos;
#endif
};
typedef struct _output_cell *output_cell;
#define OUTPUTSIZE       (long)(sizeof(struct _input_cell))
#define isoutput(p)      (type(p) == T_OUTPUT)
#define outputstatus(p)  (((output_cell)(p))->status)
#define outputindex(p)   (((output_cell)(p))->index)
#ifndef THINK_C_DA
#define outputfile(p)    (((output_cell)(p))->fp)
#else
#define outputfile(p)    (((output_cell)(p))->ref)
#define outputxpos(p)    (((output_cell)(p))->xpos)
#endif

/* c...r macros */

#define caar(p)         car(car(p))
#define cadr(p)         car(cdr(p))
#define cdar(p)         cdr(car(p))
#define cddr(p)         cdr(cdr(p))
#define cadar(p)        car(cdr(car(p)))
#define caddr(p)        car(cdr(cdr(p)))
#define cadaar(p)       car(cdr(car(car(p))))
#define cadddr(p)       car(cdr(cdr(cdr(p))))
#define cddddr(p)       cdr(cdr(cdr(cdr(p))))
#define caddddr(p)      car(cdr(cdr(cdr(cdr(p)))))
#define cdddddr(p)      cdr(cdr(cdr(cdr(cdr(p)))))

/* true or false value macro */
#ifndef USE_OLD_CONDITIONAL_VALUE
#define istrue(p)       ((p) != F)
#define isfalse(p)      ((p) == F)
#else
#define istrue(p)       ((p) != F && (p) != NIL)
#define isfalse(p)      ((p) == F || (p) == NIL)
#endif

/* control macros for Eval_Cycle */
#define s_goto(a) { eval_operator = (a); goto LOOP; }

#define s_save(a, b) {                                 \
    if (stack + 3 >= stacktop) Error("stack overflow");\
    *++stack = (b);                                    \
    *++stack = envir;                                  \
    *++stack = (pointer)(a); }

#ifndef THINK_C_DA
#define s_return(a) {                          \
    value = (a);                               \
    eval_operator = (int)*stack--;             \
    envir = *stack--;                          \
    code = *stack--;                           \
    goto LOOP; }
#else
#define s_return(a) {                          \
    value = (a);                               \
    eval_operator = (int)*stack--;             \
    envir = *stack--;                          \
    code = *stack--;                           \
    if (eval_count++ >= EVALCOUNT) return;     \
    else goto LOOP; }
#endif

#define s_push(a) {                                 \
    if (++stack>= stacktop) Error("stack overflow");\
    *stack = (a); }

#define s_pop() (*stack--)

#define clear_args() { stack = args - 2; }

#define r_push(a) {                       \
    if (++read_stack>= top_rd_stack)     \
        Error("read_stack overflow");     \
    *read_stack = (a); }

#define r_pop() (int)(*read_stack--)

#define Init_read_stack() {        \
    read_stack = btm_rd_stack; \
    read_stack[0] = OP_RDEND; }

#define g_push(a) gc_protect_list[n_gc_protect++] = (a);

#define g_pop(n) n_gc_protect -= (n);

#include "op.h"

extern long Heap_Size;

#ifdef THINK_C_DA
extern int eval_operator;
#endif

extern pointer envir;
extern pointer code;
extern pointer *stack;
extern pointer *stackbottom;
extern pointer *stacktop;
extern long Stack_Size;
extern pointer *args;
extern long nargs;
extern pointer std_input;
extern pointer cur_input;
extern pointer std_output;
extern pointer cur_output;
extern read_op_type *read_stack;
extern read_op_type *btm_rd_stack;
extern read_op_type *top_rd_stack;
extern pointer *gc_protect_list[];
extern int n_gc_protect;
extern pointer oblist[];
extern pointer global_env;
extern pointer NIL;
extern pointer T;
extern pointer F;
extern pointer eof_object;
extern pointer LAMBDA;
extern pointer QUOTE;
extern pointer QQUOTE;
extern pointer UNQUOTE;
extern pointer UNQUOTESP;
extern pointer APPLY_ARRAY;
extern pointer RESET_PATCH;

extern jmp_buf reset_jmp;
extern char gc_verbose;

extern int Is_read_complete;
extern pointer r_value;
extern char strbuff[];

#ifndef THINK_C_DA
extern int Is_need_prompt;
#endif

#ifdef THINK_C_DA
extern int eval_status;
#define EVAL_ST_EXEC          0
#define EVAL_ST_NORMAL        1
#define EVAL_ST_PR_READ     2
#define EVAL_ST_PR_READCHAR 3
#define EVAL_ST_PR_PEEKCHAR 4
#endif

/* global variable for character table */
#define CHARMAX  256
#define CHARMASK 255
#define OCTALFORMAT "#\\%03o"
extern struct _char_cell char_table[];

#define NO_NUM_TYPE    0
#define NUM_TYPE_INT   1
#define NUM_TYPE_REAL  2

extern pointer ld_file_list[];
extern int load_files;
extern pointer i_port_list[];
extern int input_ports;
extern pointer o_port_list[];
extern int output_ports;

extern char *Get_Heap();
extern pointer cons();
extern pointer mk_atom();
extern pointer mk_string();
extern pointer mk_const();
extern pointer mki_f_string();
extern pointer mki_f_number();
extern long get_longint_from_integer();
extern pointer str_to_integer();
extern pointer int_to_string();
extern pointer mk_real();
extern pointer mk_sym();
extern pointer gensym();
extern pointer mk_sym_from_cstr();
extern pointer mk_closure();
extern pointer mk_vector();
extern pointer mk_continuation();
extern pointer mk_input();
extern pointer mk_output();
extern pointer mk_box();
#ifndef THINK_C_DA
extern pointer Eval_Cycle();
#else
extern Eval_Cycle();
#endif
extern pointer numerical_operator();
extern long printlist();
extern tok_type get_token();
extern int hash();
