;; init file for AM-Scheme
;; originally written by Atsushi Moriwaki
;; revised 18-Mar-1992 by Akira Kida.
(define else #t)

(define nil #f)
(define t #t)

(define (caar x) (car (car x)))
(define (cadr x) (car (cdr x)))
(define (cdar x) (cdr (car x)))
(define (cddr x) (cdr (cdr x)))
(define (caaar x) (car (car (car x))))
(define (caadr x) (car (car (cdr x))))
(define (cadar x) (car (cdr (car x))))
(define (caddr x) (car (cdr (cdr x))))
(define (cdaar x) (cdr (car (car x))))
(define (cdadr x) (cdr (car (cdr x))))
(define (cddar x) (cdr (cdr (car x))))
(define (cdddr x) (cdr (cdr (cdr x))))
(define (caaaar x) (car (car (car (car x)))))
(define (caaadr x) (car (car (car (cdr x)))))
(define (caadar x) (car (car (cdr (car x)))))
(define (caaddr x) (car (car (cdr (cdr x)))))
(define (cadaar x) (car (cdr (car (car x)))))
(define (cadadr x) (car (cdr (car (cdr x)))))
(define (caddar x) (car (cdr (cdr (car x)))))
(define (cadddr x) (car (cdr (cdr (cdr x)))))
(define (cdaaar x) (cdr (car (car (car x)))))
(define (cdaadr x) (cdr (car (car (cdr x)))))
(define (cdadar x) (cdr (car (cdr (car x)))))
(define (cdaddr x) (cdr (car (cdr (cdr x)))))
(define (cddaar x) (cdr (cdr (car (car x)))))
(define (cddadr x) (cdr (cdr (car (cdr x)))))
(define (cdddar x) (cdr (cdr (cdr (car x)))))
(define (cddddr x) (cdr (cdr (cdr (cdr x)))))

(define (error . args)
  (display "Error:")
  (for-each (lambda (value) (display " ") (display value)) args)
  (reset))

(define (add1 x) (+ x 1))

(define 1+ add1)

(define (sub1 x) (- x 1))

(define -1+ sub1)

(define (gcd . args)
  (define (gcd-2 a b)
    (if (zero? b) a
      (gcd-2 b (remainder a b))))
  (if (null? args) 0
    (let foo ((a (abs (car args)))
              (d (cdr args)))
      (if (pair? d)
        (foo (gcd-2 a (abs (car d))) (cdr d))
        a))))

(define (lcm . args)
  (let ((product (abs (apply * args)))
        (gcd-of-args (apply gcd args)))
      (if (zero? gcd-of-args) 1
        (quotient product gcd-of-args))))

(define (call-with-input-file string proc)
  (let ((input-port (open-input-file string))
        (old-reset-patch -*-reset-patch-*-))
    (set! -*-reset-patch-*-
      (lambda ()
        (close-input-port input-port)
        (set! -*-reset-patch-*- old-reset-patch)
        (old-reset-patch)))
    (proc input-port)
    (close-input-port input-port)
    (set! -*-reset-patch-*- old-reset-patch)
    #t))

(define (call-with-output-file string proc)
  (let ((output-port (open-output-file string))
        (old-reset-patch -*-reset-patch-*-))
    (set! -*-reset-patch-*-
      (lambda ()
        (close-output-port output-port)
        (set! -*-reset-patch-*- old-reset-patch)
        (old-reset-patch)))
    (proc output-port)
    (close-output-port output-port)
    (set! -*-reset-patch-*- old-reset-patch)
    #t))

(define call/cc call-with-current-continuation)

(define (atom? x) (not (pair? x)))

(define number? real?)

(define complex? real?)

(define rational? integer?)

(define (exact->inexact x) (+ x 0.0))

(define (procedure? x) 
  (or (primitive-procedure? x)
      (closure? x)
      (continuation? x)
      (macro? x)))

(define (some? proc plist)
  (cond
    ((null? plist) #f)
    ((proc (car plist)) #t)
    (else (some? proc (cdr plist)))))

(define (every? proc plist)
  (cond
    ((null? plist) #t)
    ((proc (car plist)) (every? proc (cdr plist)))
    (else #f)))

(define (map proc arg . rest)
  (define (map1 proc list)
    (if (pair? list)
      (cons (proc (car list)) (map1 proc (cdr list)))
      '()))
  (define (maps proc dlist)
    (if (every? pair? dlist)
      (cons (apply proc (map1 car dlist))
            (maps proc (map1 cdr dlist)))
      '()))
  (if (pair? rest)
    (maps proc (cons arg rest))
    (map1 proc arg)))

(define (for-each proc . args)
  (let for-each-loop ((dlist args))
    (if (every? pair? dlist)
      (begin (apply proc (map car dlist))
             (for-each-loop (map cdr dlist)))
      #t)))

(define (last-pair x)
    (if (pair? (cdr x))
        (last-pair (cdr x))
        x))

(define (vector . x)
  (let* ((len (length x))
         (vec (make-vector len)))
        (let vec-loop ((l x) (i 0))
          (if (>= i len)
              vec
              (begin (vector-set! vec i (car l))
                     (vec-loop (cdr l) (+ i 1)))))))

(define (string . x)
  (let* ((len (length x))
         (str (make-string len)))
        (let str-loop ((l x) (i 0))
          (if (>= i len)
              str
              (begin (string-set! str i (car l))
                     (str-loop (cdr l) (+ i 1)))))))

(define (string->list s)
  (let str-loop ((l '())
                 (i (- (string-length s) 1)))
    (if (< i 0)
      l
      (str-loop
        (cons (string-ref s i) l)
        (- i 1)))))

(define (list->string x)
  (let* ((len (length x))
         (str (make-string len)))
    (let str-loop ((l x) (i 0))
      (if (>= i len)
        str
        (begin
          (string-set! str i (car l))
          (str-loop (cdr l) (+ i 1)))))))


;;
;; macro for do (don't use symbol -*-doloop-*- in the do expreesion.)
;;
;; (do ((<var1> <init1> <step1>) ...)
;;     (<test> <sequence>)
;;     <command1> ...)
;; ==>
;; (let -*-doloop-*- ((<var1 <init1>) ...)
;;   (if <test>
;;       (begin <sequence>)
;;       (begin <command1> ... (-*-doloop-*- <step1> ...)))
;;
(macro do (lambda (l)
  (define (do-macro init-step test-seq . com)
    (list 'let 
          '-*-doloop-*- 
          (let mk-init ((x init-step))
            (if (null? x)
                '()
                (cons (list (caar x) (cadar x)) (mk-init (cdr x)))))          
          (list 'if 
                (car test-seq) 
                (cons 'begin (cdr test-seq)) 
                (append (cons 'begin com) 
                        (cons (cons '-*-doloop-*- 
                                    (let mk-step ((x init-step))
                                      (if (null? x)
                                          '()
                                          (cons (if (null? (cddar x)) 
                                                    (caar x) 
                                                    (car (cddar x)))
                                                (mk-step (cdr x)))))) '())))))

  (apply do-macro (cdr l))))

;
;   Copyright 1988 by Eric S. Tiedemann; all rights reserved.
;
;
(macro
 quasiquote
 (lambda (l)
   (define (mcons f l r)
     (if (and (pair? r)
              (eq? (car r) 'quote)
              (eq? (cadr r) (cdr f))
              (pair? l)
              (eq? (car l) 'quote)
              (eq? (cadr l) (car f)))
         (list 'quote f)
         (list 'cons l r)))
   (define (mappend f l r)
     (if (or (null? (cdr f))
             (and (pair? r)
                  (eq? (car r) 'quote)
                  (eq? (cadr r) '())))
        l
         (list 'append l r)))
   (define (mvec form v v1)
     (if (and (pair? v1) (eq? (car v1) 'quote) (eq? (cadr v1) v))
         form
         (list 'list->vector v1)))
   (define (foo level form)
     (cond ((vector? form)
            (let ((v (vector->list form)))
              (mvec form v (foo level v))))
           ((atom? form) (list 'quote form))
           ((eq? 'quasiquote (car form))
            (mcons form ''quasiquote (foo (add1 level) (cdr form))))
           (#t (if (zero? level)
                   (cond ((eq? (car form) 'unquote) (cadr form))
                         ((eq? (car form) 'unquote-splicing)
                          (error "Unquote-splicing wasn't in a list:"
                                 form))
                         ((and (pair? (car form))
                               (eq? (caar form) 'unquote-splicing))
                          (mappend form (cadar form) (foo level (cdr
                                                                 form))))
                         (#t (mcons form (foo level (car form))
                                         (foo level (cdr form)))))
                   (cond ((eq? (car form) 'unquote)
                          (mcons form ''unquote (foo (sub1 level)
                                                     (cdr form))))
                         ((eq? (car form) 'unquote-splicing)
                          (mcons form ''unquote-splicing
                                      (foo (sub1 level) (cdr form))))
                         (#t (mcons form (foo level (car form))
                                         (foo level (cdr form)))))))))
   (foo 0 (cadr l))))
