/*
 * eval.c
 *
 * This file is part of AM-Scheme.
 * AM-Scheme is coded by Atsushi Moriwaki
 *
 * This software is completely free to copy and distribute modifed version.
 * But I would appreciate it if you left my name on the code as the author.
 *
 */

#include "amscm.h"

#ifdef THINK_C_DA
int eval_operator;
pointer value;
int eval_status = EVAL_ST_NORMAL;
#endif

/* ========== Evaluation Cycle ========== */

errorArgs(pName, supplied, required)	/* by Akira Kida */
char *pName;
long supplied;
int required;
{
    if (required < 0)
        Error("'%s' -- %ld arguments, but at least %d is required",
	        pName, supplied - 1, -required);
    else if (supplied < required)
        Error("'%s' -- %ld arguments, while %d is required",
	        pName, supplied - 1, required - 1);
    else
        Error("'%s' -- %ld arguments, only %d is sufficient",
	        pName, supplied - 1, required - 1);

}

/* kernel of this intepreter */
#ifndef THINK_C_DA
pointer Eval_Cycle(eval_operator)
register int eval_operator;
#else
Eval_Cycle()
#endif
{
#ifndef THINK_C_DA
    register pointer value, x, y;
#else
    register pointer x, y;
    register int eval_count = 0;
    extern int verbose_eval;
#endif
    register long v, i;
    int pArgs;
    pointer g;

LOOP:
    switch(eval_operator) {
case OP_EVALEND: /* print evalution result */
#ifndef THINK_C_DA
    return (value);
#else
    if (verbose_eval) {
        if (outputxpos(std_output)) PrintChar(std_output, '\n');
        PrintString(std_output, "==> ");
        printlist(value, std_output, 1, 1);
        PrintChar(std_output, '\n');
    }
    eval_status = EVAL_ST_NORMAL;
    r_value = NIL;
    return;
#endif
    
case OP_EVAL: /* main part of evalution */
    if (issymbol(code)) { /* symbol */
        for ( x = envir ; x != NIL ; x = cdr(x)) {
            for (y = car(x) ; y != NIL ; y = cdr(y))
                if (caar(y) == code) break;
            if (y != NIL) break;
        }
        if (x != NIL) s_return(cdar(y))
        else {
            Error("Unbound variable %s", symname(code));
        }
    } else if (ispair(code)) {
        if (issyntax(x = car(code))) { /* SYNTAX */
            code = cdr(code);
            s_goto(syntaxnum(x))
        } else { /* first, eval top element and eval arguments */
            s_save(OP_E0ARGS, code)
            code = car(code);
            s_goto(OP_EVAL)
        }
    } else s_return(code)

case OP_E0ARGS: /* eval arguments */
    if (ismacro(value)) { /* macro expansion */
        s_save(OP_DOMACRO, NIL)
        s_push(NULL) s_push(value) s_push(code)
        args = stack -1;
        nargs = 2;
        s_goto(OP_APPLY)
    } else {
        s_push(NULL)
        code = cdr(code); 
        s_goto(OP_E1ARGS) 
    }

case OP_E1ARGS: /* eval arguments */
    s_push(value)
    if (ispair(code)) { /* continue */
        s_save(OP_E1ARGS, cdr(code))
        code = car(code);
        s_goto(OP_EVAL)
    } else if (issymbol(code)) { /* extended */
        s_save(OP_E2ARGS, NIL)
        s_goto(OP_EVAL)
    } else
        s_goto(OP_E3ARGS)

case OP_E2ARGS:     /* extended proc call syntax, added by Akira Kida */
    if (ispair(value))
        s_push(car(value))
    code = NIL;
    s_goto(OP_E3ARGS)

case OP_E3ARGS:  /* by Akira Kida */
    {
        register pointer *s = stack;
        nargs = 0;
        while (*s != NULL) { ++nargs; --s; }
        args = ++s;
        s_goto(OP_APPLY)
    }
    
case OP_APPLY: /* apply 'code' to 'args' */
    switch(type(args[0])) {
    case T_PRIM1: /* PRIM TYPE 1 */
	if ((pArgs = primargs(args[0])) > 0) {
            if (pArgs != nargs)
                     errorArgs(primname(args[0]), nargs, pArgs);
	} else {
             if ((1 - pArgs) > nargs)
                     errorArgs(primname(args[0]), nargs, pArgs);
	}
        s_goto(primnum(args[0]))
    
    case T_PRIM2: /* PRIM TYPE 2 */
	if ((pArgs = primargs(args[0])) > 0) {
            if (pArgs != nargs)
                     errorArgs(primname(args[0]), nargs, pArgs);
	} else {
             if ((1 - pArgs) > nargs)
                     errorArgs(primname(args[0]), nargs, pArgs);
	}
        x = primproc(args[0])();
        clear_args();
        s_return(x);
    
    case T_CLOSURE: /* CLOSURE */
    case T_MACRO: /* MACRO */
        /* make environment */
        envir = cons(NIL, closure_env(args[0]));
        g_push(&g) 
        for (g = car(closure_code(args[0])), v = 1;
             ispair(g) ; g = cdr(g), ++v) {
            if (v >= nargs) Error("closure -- few arguments");
            else  {
                y = cons(car(g), args[v]);
                car(envir) = cons(y, car(envir));
            }
        }
        if (g == NIL) {
            if (v < nargs) Error("closure -- many arguments");
        } else if (issymbol(g)) {
            i = nargs - 1;
            for (y = NIL ; i >= v; --i)
                y = cons(args[i], y);
            y = cons(g, y);
            car(envir) = cons(y, car(envir));
        } else Error("closure -- syntax error");
        code = cdr(closure_code(args[0]));
        g_pop(1)
        clear_args()
        s_goto(OP_BEGIN)
     
     case T_CONTINUATION: /* CONTINUATION */
     {  
        pointer *copy = cont_copy(args[0]);
        long len = cont_len(args[0]);
        x = (nargs >= 2) ? args[1] : NIL;
        for(stack = stackbottom - 1; len != 0 ; --len)
            *++stack = *copy++;
        s_return(x)
     }
     
     default:
         Error("Illegal function");
     }

case OP_DOMACRO: /* do macro */
    code = value;
    s_goto(OP_EVAL)

case OP_LAMBDA: /* lambda */
    s_return(mk_closure(code, envir))

case OP_QUOTE: /* quote */
    s_return(car(code))

case OP_DEF0: /* define */
    if (ispair(car(code))) {
        g = caar(code);
        g_push(&g)
        x = cons(cdar(code), cdr(code));
        code = cons(LAMBDA, x);
        g_pop(1)
    } else {
        g = car(code);
        code = cadr(code);
    }
    if (!issymbol(g)) Error("define -- variable is not symbol");
    s_save(OP_DEF1, g)
    s_goto(OP_EVAL)

case OP_DEF1: /* define */
    for (x = car(envir) ; x != NIL ; x = cdr(x))
        if (caar(x) == code) break;
    if (x != NIL) cdar(x) = value;
    else {
        x = cons(code, value);
        car(envir) = cons(x, car(envir));
    }
    s_return(code)

case OP_GDEF0: /* define-global */
    if (ispair(car(code))) {
        g = caar(code);
        g_push(&g)
        x = cons(cdar(code), cdr(code));
        code = cons(LAMBDA, x);
        g_pop(1)
    } else {
        g = car(code);
        code = cadr(code);
    }
    if (!issymbol(g)) Error("define-global -- variable is not symbol");
    s_save(OP_GDEF1, g)
    s_goto(OP_EVAL)

case OP_GDEF1: /* define-global */
    for (x = car(global_env) ; x != NIL ; x = cdr(x))
        if (caar(x) == code) break;
    if (x != NIL) cdar(x) = value;
    else {
        x = cons(code, value);
        car(global_env) = cons(x, car(global_env));
    }
    s_return(code)
       
case OP_SET0: /* set! */ 
    s_save(OP_SET1, car(code))
    code = cadr(code);
    s_goto(OP_EVAL)

case OP_SET1: 
    for ( x = envir ; x != NIL ; x = cdr(x)) {
        for (y = car(x) ; y != NIL ; y = cdr(y))
            if (caar(y) == code) break;
        if (y != NIL) break;
    }
    if (x != NIL) cdar(y) = value;
    else Error("Unbound variable %s", symname(code));
        
case OP_BEGIN: /* begin */
    if (!ispair(code)) s_return(code)
    if (cdr(code) != NIL) s_save(OP_BEGIN, cdr(code))
    code = car(code);
    s_goto(OP_EVAL)

case OP_IF0: /* if */
    s_save(OP_IF1, cdr(code))
    code = car(code);
    s_goto(OP_EVAL)

case OP_IF1: /* if */
    if (istrue(value)) code = car(code);
    else code = cadr(code); /* (if #f 1) ==> () because car(NIL) = NIL */
    s_goto(OP_EVAL)

case OP_COND0: /* cond */
    if (!ispair(code)) Error("cond -- syntax error");
    s_save(OP_COND1, code)
    code = caar(code);
    s_goto(OP_EVAL)

case OP_COND1: /* cond */
    if (istrue(value)) {
        if ((code = cdar(code)) == NIL) s_return(value)
        else {
            if (car(code) == APPLY_ARRAY) {
                s_save(OP_COND2, value)
                code = cadr(code);
                s_goto(OP_EVAL)
            } else s_goto(OP_BEGIN)
        }
    } else {
        if ((code = cdr(code)) == NIL) s_return(NIL)
        else {
            s_save(OP_COND1, code)
            code = caar(code);
            s_goto(OP_EVAL)
        }
    }

case OP_COND2: /* cond */
    s_push(NULL) s_push(value) s_push(code)
    args = stack - 1; nargs = 2;
    s_goto(OP_APPLY)
    
case OP_LET0: /* let */
    s_push(NULL)
    value = code;
    code = issymbol(car(code)) ? cadr(code) : car(code);
    s_goto(OP_LET1)

case OP_LET1: /* let (caluculate parameters) */
    s_push(value);
    if (ispair(code)) { /* continue */
        s_save(OP_LET1, cdr(code))
        code = cadar(code);
        s_goto(OP_EVAL)
    } else { /* end */
        register pointer *s = stack;
        nargs = 0;
        while (*s != NULL) { ++nargs; --s; }
        args = ++s;
        envir = cons(NIL, envir);
        g_push(&g)
        for (g = issymbol(car(args[0])) ? cadr(args[0]) : car(args[0]), i = 1 ;
             i < nargs ; g = cdr(g), i++) {
            y = cons(caar(g), args[i]);
            car(envir) = cons(y, car(envir));
        }
        g_pop(1)
        if (issymbol(car(args[0]))) { /* named let */
            g_push(&g)
            for (g = car(envir), y = NIL ; g != NIL ; g = cdr(g))
                y = cons(caar(g), y);
            g_pop(1)
            x = cons(y, cddr(args[0]));
            x = mk_closure(x, envir);
            x = cons(car(args[0]), x);
            car(envir) = cons(x, car(envir));
            code = cddr(args[0]);
        } else code = cdr(args[0]);
        clear_args()
        s_goto(OP_BEGIN)
    }

case OP_LAST0: /* let* -- debugged by Akira Kida */
    if (car(code) == NIL) {
        envir = cons(NIL, envir);
        code = cdr(code);
        s_goto(OP_BEGIN)
    }
    s_push(cdr(code))
    s_save(OP_LAST1, car(code))
    code = cadaar(code);
    s_goto(OP_EVAL)

case OP_LAST1: /* let* (make new frame) */
    envir = cons(NIL, envir);
    s_goto(OP_LAST2)

case OP_LAST2: /* let* (calculate parameters) */
    x = cons(caar(code), value);
    car(envir) = cons(x, car(envir));
    code = cdr(code);
    if (ispair(code)) { /* continue */
        s_save(OP_LAST2, code)
        code = cadar(code);
        s_goto(OP_EVAL)
    } else { /* end */
        code = s_pop();
        s_goto(OP_BEGIN)
    }

case OP_LREC0: /* letrec */
    envir = cons(NIL, envir);
    s_push(NULL)
    value = code;
    code = car(code);
    s_goto(OP_LREC1)

case OP_LREC1: /* letrec (caluculate parameters) */
    s_push(value);
    if (ispair(code)) { /* continue */
        s_save(OP_LREC1, cdr(code))
        code = cadar(code);
        s_goto(OP_EVAL)
    } else { /* end */
        register pointer *s = stack;
        nargs = 0;
        while (*s != NULL) { ++nargs; --s; }
        args = ++s;
        g_push(&g)
        for (g = car(args[0]),
             i = 1 ;  i < nargs ; g = cdr(g), i++) {
            x = cons(caar(g), args[i]);
            car(envir) = cons(x, car(envir));
        }
        g_pop(1)
        code = cdr(args[0]); 
        clear_args()
        s_goto(OP_BEGIN)
    }

case OP_MAC0: /* macro */
    x = car(code);
    code = cadr(code);
    if (!issymbol(x)) Error("macro -- variable is not symbol");
    s_save(OP_MAC1, x)
    s_goto(OP_EVAL)

case OP_MAC1: /* macro */
    type(value) = T_MACRO;
    for (x = car(envir) ; x != NIL ; x = cdr(x))
        if (caar(x) == code) break;
    if (x != NIL) cdar(x) = value;
    else {
        y = cons(code, value);
        car(envir) = cons(y, car(envir));
    }
    s_return(code)

case OP_GMAC0: /* macro-global */
    x = car(code);
    code = cadr(code);
    if (!issymbol(x)) Error("macro-global -- variable is not symbol");
    s_save(OP_GMAC1, x)
    s_goto(OP_EVAL)

case OP_GMAC1: /* macro-global */
    type(value) = T_MACRO;
    for (x = car(global_env) ; x != NIL ; x = cdr(x))
        if (caar(x) == code) break;
    if (x != NIL) cdar(x) = value;
    else {
        y = cons(code, value);
        car(global_env) = cons(y, car(global_env));
    }
    s_return(code)

case OP_AND0: /* and */
    if (code == NIL) s_return(T)
    s_save(OP_AND1, cdr(code))
    code = car(code);
    s_goto(OP_EVAL)

case OP_AND1: /* and */
    if (isfalse(value)) s_return(value)
    else if (code == NIL) s_return(value)
    else {
        s_save(OP_AND1, cdr(code))
        code = car(code);
        s_goto(OP_EVAL)
    }

case OP_OR0: /* or */
    if (code == NIL) s_return(F)
    s_save(OP_OR1, cdr(code))
    code = car(code);
    s_goto(OP_EVAL)

case OP_OR1: /* or */
    if (istrue(value)) s_return(value)
    else if (code == NIL) s_return(value)
    else {
        s_save(OP_OR1, cdr(code))
        code = car(code);
        s_goto(OP_EVAL)
    }

case OP_CASE0: /* case */
    s_save(OP_CASE1, cdr(code))
    code = car(code);
    s_goto(OP_EVAL)

case OP_CASE1: /* case */
    for (x = code; x != NIL; x = cdr(x)) {
        if (!ispair(y = caar(x))) break;
        for ( ; y != NIL; y = cdr(y))
            if (eqv(car(y), value)) break;
        if (y != NIL) break;
    }
    if (x != NIL) {
        if (ispair(caar(x))) {
            code = cdar(x);
            s_goto(OP_BEGIN)
        } else {/* else */
            s_save(OP_CASE2, cdar(x))
            code = caar(x);
            s_goto(OP_EVAL)
        }
    } else s_return(NIL)

case OP_CASE2: /* case */
    if (istrue(value)) s_goto(OP_BEGIN)
    else s_return(NIL)

case OP_PRIMEVAL: /* eval */
    code = args[1];
    if (nargs >= 3) envir = args[2];
    clear_args()
    s_goto(OP_EVAL)

case OP_PRIMAPPLY: /* apply */
    for (i = 1 ;  i < nargs - 1 ; i++)
        args[i-1] = args[i];
    stack = &args[nargs-3];
    for (x = args[nargs-1], nargs -= 2; ispair(x) ; x = cdr(x), ++nargs)
        s_push(car(x));
    s_goto(OP_APPLY)

case OP_CONTINUATION: /* call/cc */
    args[0] = args[1];
    args[1] = mk_continuation(args - 2);
    stack = &args[1];
    s_goto(OP_APPLY)

case OP_LOAD: /* load */
    if (!isstring(args[1])) Error("load -- arg is not string");
#ifndef THINK_C_DA
    clear_args()
    LoadFromFileName(strvalue(args[1]), 1, 0);
    s_return (T)
#else
    clear_args()
    LoadFromFileName(strvalue(args[1]));
    eval_status = EVAL_ST_NORMAL;
    return;
#endif

case OP_READ: /* read */
#ifndef THINK_C_DA
{
    tok_type tok;
    
    if (nargs >= 2) {
        if (!isinput(x = args[1]))
            Error("read -- second arg is not input port");
    } else x = cur_input;
    Is_need_prompt = 0;
    for (;;) {
        tok = get_token(x);
        Is_read_complete = 0;
        Read_Cycle(tok);
        if (Is_read_complete) {
            Init_read_stack()
            read_white(x);
            clear_args()
            Is_need_prompt = 1;
            s_return (r_value);
        }
    }
}
#else
    if (nargs >= 2) {
        if (!isinput(x = args[1]))
            Error("read -- second arg is not input port");
    } else x = cur_input;
    if (x == std_input) {
        clear_args()
        eval_status = EVAL_ST_PR_READ;
        return;
    }
    if (inputstatus(x) == 0)
       Error("read -- input port is not active");        
    Read_From_Input(x); /* eval_status is EVAL_ST_EXEC */
    clear_args()
    s_return (r_value)
#endif

case OP_READ1: /* read */
    s_return (r_value)

case OP_RCHR: /* read-char */
#ifndef THINK_C_DA
{
    int ch;
    
    if (nargs >= 2) {
        if (!isinput(x = args[1]))
            Error("read-char -- arg is not input port");
    } else x = cur_input;
    if (inputstatus(x) == 0)
       Error("read-char -- input port is not active");
    Is_need_prompt = 0;
    ch = read_char(x);
    Is_need_prompt = 1;
    clear_args()
    if (ch == EOF) s_return(eof_object)
    else s_return((pointer)&char_table[ch])       
}
#else
{
    int ch;
    
    if (nargs >= 2) {
        if (!isinput(x = args[1]))
            Error("read-char -- arg is not input port");
    } else x = cur_input;
    if (x == std_input) s_goto(OP_RCHR1)
    if (inputstatus(x) == 0)
       Error("read-char -- input port is not active");
    ch = read_char(x);
    clear_args()
    if (ch == EOF) s_return(eof_object)
    else s_return((pointer)&char_table[ch])       
}
#endif
    
case OP_RCHR1: /* read-char */
#ifdef THINK_C_DA
{
    clear_args()
    if (inputpos(std_input) >= inputeol(std_input)) {
        eval_status = EVAL_ST_PR_READCHAR;
        clear_inputline(std_input)
        return;
    } else {
       int ch = read_char(std_input);
       s_return((pointer)&char_table[ch])
   }
}
#endif

case OP_PKCH: /* peek-char */
#ifndef THINK_C_DA
{
    int ch;
    
    if (nargs >= 2) {
        if (!isinput(x = args[1]))
            Error("peek-char -- second arg is not input port");
    } else x = cur_input;
    if (inputstatus(x) == 0)
       Error("peek-char -- input port is not active");
    Is_need_prompt = 0;
    ch = read_char(x);
    back_char(x);
    Is_need_prompt = 1;
    clear_args()
    if (ch == EOF) s_return(eof_object)
    else s_return((pointer)&char_table[ch])       
}
#else
{
    int ch;
    
    if (nargs >= 2) {
        if (!isinput(x = args[1]))
            Error("peek-char -- second arg is not input port");
    } else x = cur_input;
    if (x == std_input) s_goto(OP_PKCH1)
    if (inputstatus(x) == 0)
       Error("peek-char -- input port is not active");
    ch = read_char(x);
    back_char(x);
    clear_args()
    if (ch == EOF) s_return(eof_object)
    else s_return((pointer)&char_table[ch])       
}
#endif
    
case OP_PKCH1: /* peek-char */
#ifdef THINK_C_DA
{
    clear_args()
    if (inputpos(std_input) >= inputeol(std_input)) {
        eval_status = EVAL_ST_PR_PEEKCHAR;
        clear_inputline(std_input)
        return;
    } else {
       int ch = read_char(std_input);
       back_char(std_input);
       s_return((pointer)&char_table[ch])
   }
}
#endif

default:
    Error("Illegal oprerator %d", eval_operator);
    }
}
