/*
 * generic_os.c
 *
 * This file is part of AM-Scheme.
 * AM-Scheme is coded by Atsushi Moriwaki
 *
 * This software is completely free to copy and distribute modifed version.
 * But I would appreciate it if you left my name on the code as the author.
 *
 */

#include "amscm.h"

int Is_need_prompt = 1;

PrintChar(f, ch)
pointer f;
char ch;
{
    putc(ch, outputfile(f));
    if (ch == '\n') fflush(outputfile(f));
}

PrintString(f, str)
pointer f;
char *str;
{
    char ch;
    
    while ((ch = *str++) != '\0') PrintChar(f, ch);
}

PrStrWithLen(fp, str, len)
pointer fp;
char *str;
long len;
{
    long i;

    for (i = 0 ; i < len ; i++) PrintChar(fp, str[i]);
}


/*
 *	Get a line using DOS "Buffered Keyboard Input" system call,
 *	so as to enable some history-shell functionalities if any.
 *	This function is added by Akira Kida.
 */
#if defined(MSDOS)
#define	KBD_EOF	'\x1a'
char *
historyGets(buf, size, fp)
char *buf;
unsigned int size;
FILE *fp;
{
	char	_far *bufptr;	/* MS-DOS only */
	union REGS	regs;
	struct SREGS	sRegs;
        void (*sigproc)();	/* void is Ok, becaus this is DOS specific */

	if (!isatty(fileno(fp)))	/* fp is redirected to a file */			return fgets(buf, size, fp);
	/* fake input buffer */
	buf[0] = (char)(size - 3 > 0xff ? 0xff : size - 3);
	bufptr = (char _far*)buf;
	segread(&sRegs);
	sRegs.ds = FP_SEG(bufptr);
	regs.x.dx = FP_OFF(bufptr);
	regs.h.ah = 0x0a;
	sigproc = signal(SIGINT, SIG_IGN);
	intdosx(&regs, &regs, &sRegs);
	signal(SIGINT, sigproc);
	putchar('\n');
	strcpy(buf + (unsigned char)buf[1] + 2, "\n");
	memmove(buf, buf + 2, strlen(buf) + 1);
	if (*buf == KBD_EOF)
		return	NULL;
	return	buf;
}
#else
/*
 * In non DOS envrionment, we simply call ``fgets''.
 */
# define historyGets(buf, size, fp)	fgets((buf), (size), (fp))
#endif

Get_Line(l)
pointer l;
{
    clear_inputline(l);
    if (l == std_input && Is_need_prompt)
        PrintString(std_output, prompt);
    if (historyGets(inputbuffer(l), LINESIZE, inputfile(l)) != NULL) {
        inputpos(l) = 0;
        inputeol(l) = strlen(inputbuffer(l));
        return (1);
    } else {
        return (EOF);
    }
}


LoadFromFileName(name, IsVerbose, IsInit)
char *name;
int IsVerbose;
int IsInit;
{
    FILE *fp;
    pointer l;

    if ((fp = fopen(name, "r")) == NULL) {
        if (IsInit) {
            sprintf(&strbuff[0], "Warnning: Unable to open %s\n", name);
            PrintString(std_output, &strbuff[0]);
            return 0;
        } else Error("Unable to open %s", name);
    }
    if (load_files >= MAXFILE) {
        fclose(fp);
        Error("Too many open files");
    }
    l = mk_input(fp, 1);
    ld_file_list[load_files++] = l;
    sprintf(&strbuff[0], ";loading %s\n", name);
    PrintString(std_output, &strbuff[0]);
    Rep_From_Input(l, IsVerbose);
    fclose(fp);
    --load_files;
    Init_read_stack()
    return 1;
}

Rep_From_Input(l, IsVerbose)
pointer l;
int IsVerbose;
{
    tok_type tok;
    
    g_push(&l)
    for(;;) {
        tok = get_token(l);
        Is_read_complete = 0;
        Read_Cycle(tok);
	Is_need_prompt = 0;
        if (Is_read_complete) {
            Init_read_stack()
	    Is_need_prompt = 1;
            if (r_value == eof_object) {
                g_pop(1)
                break;
            }
            read_white(l);
            s_save(OP_EVALEND, code)
            code = r_value;
            r_value = Eval_Cycle(OP_EVAL);
            if (IsVerbose) {
                printlist(r_value, std_output, 1, 1);
                PrintChar(std_output, '\n');
            }
        }
    }
    return 1;
}

#ifdef THINK_C
#include <console.h>
#endif

main(argc, argv)
int argc;
char *argv[];
{
    int ret;

#ifdef THINK_C
    argc = ccommand(&argv);
#endif

    /* get arguments */
    while (--argc) {
        if (**++argv != '-')
            break;
        else {
            switch ((*argv)[1]) {
            case 'h' : Heap_Size = atol(&(*argv)[2]);
                       break;
            case 's' : Stack_Size = atol(&(*argv)[2]);
                       break;
            default  : break;
            }
        }
    }
   
    printf(Banner);
    if (init_scheme())
        fprintf(stderr, "Initialization Error\n");
    else if (setjmp(reset_jmp) == 0) {
        /* changed by A.Kida */
        LoadFromFileName(InitFileName, 0, 1);
        /* A detection of any error in a file which is specified as a
         * command line parameter aborts loading and ignores the files
         * rest (if any). In such a case, AM-Scheme will begin read /
         * eval / print loop immediately.
         */
        if ((ret = setjmp(reset_jmp)) == 0)
            while (argc--)
                LoadFromFileName(*argv++, 0, 1);
        /* main loop */
        if (ret < 2)
            if (setjmp(reset_jmp) < 2)
                Rep_From_Input(std_input, 1);
    }
    exit_scheme();
    printf(GoodBye);
}

reset_io()
{
    int i;
    for (i = 0 ; i < input_ports ; i++) {
        inputstatus(i_port_list[i]) = 0;
        fclose(inputfile(i_port_list[i]));
    }
    for (i = 0 ; i < output_ports ; i++) {
        outputstatus(o_port_list[i]) = 0;
        fclose(outputfile(o_port_list[i]));
    }
    input_ports = output_ports = 0;
    
}

reset_load()
{
    int i;
    for (i = 0 ; i < load_files ; i++)
        fclose(inputfile(ld_file_list[i]));
    load_files = 0;
}

reset_os_depend() { Is_need_prompt = 1; }

exit_os_depend() {}
