/*
 * init.c
 *
 * This file is part of AM-Scheme.
 * AM-Scheme is coded by Atsushi Moriwaki
 *
 * This software is completely free to copy and distribute modifed version.
 * But I would appreciate it if you left my name on the code as the author.
 *
 */

#include "amscm.h"
#include "prim.h"

int scheme_alive = 0;

struct _cons_cell _NIL;
struct _cons_cell _T;
struct _cons_cell _F;
struct _cons_cell _eof_object;

struct _input_cell _std_input;
struct _output_cell _std_output;

extern char *Heap1, *Heap2, *Current_Heap;

extern int UpperCase();

/* ========== Initialization of internal keywords ========== */

mk_syntax(op, name)
int op;
char *name;
{
    pointer x;
    int hashpos;
    long len = strlen(name), i;
    char *new;

    strcpy(&strbuff[0], name);
    name = &strbuff[0]; 
    if (STDCHARCASE > 0) {
        /* change to upper characters */
        for (i = 0 ; i < len ; ++i) name[i] = (char)UpperCase((int)name[i]);
    }
    hashpos = hash(name, len);
    x = (pointer)Get_Heap(SYMBOLSIZE(len));
    for(i = 0, new = syntaxname(x) ; i < len ; i++) new[i] = name[i];
    new[len] = '\0';
    type(x) = T_SYNTAX;
    syntaxlen(x) = len;
    syntaxprop(x) = NIL;
    syntaxnum(x) = op;
    oblist[hashpos] = cons(x, oblist[hashpos]);
}

mk_prim1(op, name, n)
int op;
char *name;
int n;
{
    pointer x, y;

    x = mk_sym_from_cstr(name, STDCHARCASE);
    y = (pointer)Get_Heap(PRIMITIVESIZE(strlen(name)));
    strcpy(primname(y), name);
    type(y) = T_PRIM1;
    primargs(y) = n;
    primnum(y) = op;
    car(global_env) = cons(cons(x, y), car(global_env));
}

mk_prim2(proc, name, n)
pointer (*proc)();
char *name;
int n;
{
    pointer x, y;

    x = mk_sym_from_cstr(name, STDCHARCASE);
    y = (pointer)Get_Heap(PRIMITIVESIZE(strlen(name)));
    strcpy(primname(y), name);
    type(y) = T_PRIM2;
    primargs(y) = n;
    primproc(y) = proc;
    car(global_env) = cons(cons(x, y), car(global_env));
}

/*
 * initialize several globals
 *   -- modified by A.Kida
 */
init_globals()
{
    register int i;
    register pointer x;
    
    /* initialize NIL */
    type(NIL) = 0;
    car(NIL) = cdr(NIL) = NIL;
    /* init T */
    type(T) = 0;;
    car(T) = cdr(T) = T;
    /* init F */
    type(F) = 0;
    car(F) = cdr(F) = F;
    /* init eof_object */
    type(eof_object) = 0;
    car(eof_object) = cdr(eof_object) = eof_object;    
    /* init global_env */
    global_env = cons(NIL, NIL);

    /* initialize character table */
    for (i = 0 ; i < CHARMAX ; i++) {
        x = (pointer)&char_table[i];
        type(x) = T_CHAR;
        cvalue(x) = (uShort)i;
    }

    /* initialize  syntax */
    mk_syntax(OP_LAMBDA, "lambda");
    mk_syntax(OP_QUOTE, "quote");
    mk_syntax(OP_DEF0, "define");
    mk_syntax(OP_GDEF0, "define-global");
    mk_syntax(OP_IF0, "if");
    mk_syntax(OP_BEGIN, "begin");
    mk_syntax(OP_SET0, "set!");
    mk_syntax(OP_COND0, "cond");
    mk_syntax(OP_MAC0, "macro");
    mk_syntax(OP_GMAC0, "macro-global");
    mk_syntax(OP_AND0, "and");
    mk_syntax(OP_OR0, "or");
    mk_syntax(OP_LET0, "let");
    mk_syntax(OP_LREC0, "letrec");
    mk_syntax(OP_LAST0, "let*");
    mk_syntax(OP_CASE0, "case");

    /* initialize primitive procedure */
    mk_prim1(OP_PRIMEVAL, "eval", MUST_BE_N(1));
    mk_prim1(OP_PRIMAPPLY, "apply", AT_LEAST_N(2));
    mk_prim1(OP_CONTINUATION, "call-with-current-continuation", MUST_BE_N(1));
    mk_prim1(OP_LOAD, "load", MUST_BE_N(1));
    mk_prim1(OP_READ, "read", AT_LEAST_N(0));
    mk_prim1(OP_RCHR, "read-char", AT_LEAST_N(0));
    mk_prim1(OP_PKCH, "peek-char", AT_LEAST_N(0));

    mk_prim2(PR_LIST, "list", AT_LEAST_N(0));
    mk_prim2(PR_CAR, "car", MUST_BE_N(1));
    mk_prim2(PR_CDR, "cdr", MUST_BE_N(1));
    mk_prim2(PR_CONS, "cons", MUST_BE_N(2));
    mk_prim2(PR_APPEND, "append", AT_LEAST_N(0));
    mk_prim2(PR_REVERSE, "reverse", MUST_BE_N(1));
    mk_prim2(PR_LISTREF, "list-ref", MUST_BE_N(2));
    mk_prim2(PR_SETCAR, "set-car!", MUST_BE_N(2));
    mk_prim2(PR_SETCDR, "set-cdr!", MUST_BE_N(2));
    mk_prim2(PR_ADD, "+", AT_LEAST_N(0));
    mk_prim2(PR_SUB, "-", AT_LEAST_N(1));
    mk_prim2(PR_MUL, "*", AT_LEAST_N(0));
    mk_prim2(PR_QUO, "quotient", MUST_BE_N(2));
    mk_prim2(PR_REM, "remainder", MUST_BE_N(2));
    mk_prim2(PR_MOD, "modulo", MUST_BE_N(2));
    mk_prim2(PR_DIV, "/", AT_LEAST_N(1));
    mk_prim2(PR_ABS, "abs", MUST_BE_N(1));
    mk_prim2(PR_MAX, "max", AT_LEAST_N(1));
    mk_prim2(PR_MIN, "min", AT_LEAST_N(1));
    mk_prim2(PR_EVEN, "even?", MUST_BE_N(1));
    mk_prim2(PR_ODD, "odd?", MUST_BE_N(1));
    mk_prim2(PR_EXP, "exp", MUST_BE_N(1));
    mk_prim2(PR_LOG, "log", MUST_BE_N(1));
    mk_prim2(PR_SIN, "sin", MUST_BE_N(1));
    mk_prim2(PR_COS, "cos", MUST_BE_N(1));
    mk_prim2(PR_TAN, "tan", MUST_BE_N(1));
    mk_prim2(PR_ASIN, "asin", MUST_BE_N(1));
    mk_prim2(PR_ACOS, "acos", MUST_BE_N(1));
    mk_prim2(PR_ATAN, "atan", AT_LEAST_N(1));
    mk_prim2(PR_SQRT, "sqrt", MUST_BE_N(1));
    mk_prim2(PR_EXPT, "expt", MUST_BE_N(2));
    mk_prim2(PR_FLOOR, "floor", MUST_BE_N(1));
    mk_prim2(PR_CEIL, "ceiling", MUST_BE_N(1));
    mk_prim2(PR_TRUN, "truncate", MUST_BE_N(1));
    mk_prim2(PR_ROUND, "round", MUST_BE_N(1));
    mk_prim2(PR_NOT, "not", MUST_BE_N(1));
    mk_prim2(PR_BOOLP,"boolean?", MUST_BE_N(1));
    mk_prim2(PR_SYMBOLP, "symbol?", MUST_BE_N(1));
    mk_prim2(PR_INTEGERP, "integer?", MUST_BE_N(1));
    mk_prim2(PR_REALP, "real?", MUST_BE_N(1));
    mk_prim2(PR_CHP, "char?", MUST_BE_N(1));
    mk_prim2(PR_STRINGP, "string?", MUST_BE_N(1));
    mk_prim2(PR_PRIMP, "primitive-procedure?", MUST_BE_N(1));
    mk_prim2(PR_CLOSUREP, "closure?", MUST_BE_N(1));
    mk_prim2(PR_CONTP, "continuation?", MUST_BE_N(1));
    mk_prim2(PR_MACROP, "macro?", MUST_BE_N(1));
    mk_prim2(PR_PAIRP, "pair?", MUST_BE_N(1));
    mk_prim2(PR_LISTP, "list?", MUST_BE_N(1));
    mk_prim2(PR_EQV, "eqv?", MUST_BE_N(2));
    mk_prim2(PR_EQ, "eq?", MUST_BE_N(2));
    mk_prim2(PR_EQUAL, "equal?", MUST_BE_N(2));
    mk_prim2(PR_NULLP, "null?", MUST_BE_N(1));
    mk_prim2(PR_ZEROP, "zero?", MUST_BE_N(1));
    mk_prim2(PR_POSP, "positive?", MUST_BE_N(1));
    mk_prim2(PR_NEGP, "negative?", MUST_BE_N(1));
    mk_prim2(PR_NEQ, "=", AT_LEAST_N(2));
    mk_prim2(PR_LESS, "<", AT_LEAST_N(2));
    mk_prim2(PR_GRE, ">", AT_LEAST_N(2));
    mk_prim2(PR_LEQ, "<=", AT_LEAST_N(2));
    mk_prim2(PR_GEQ, ">=", AT_LEAST_N(2));
    mk_prim2(PR_WRITE, "write", AT_LEAST_N(1));
    mk_prim2(PR_DISPLAY, "display", AT_LEAST_N(1));
    mk_prim2(PR_NEWLINE, "newline", AT_LEAST_N(0));
    mk_prim2(PR_WRCHAR, "write-char", AT_LEAST_N(1));
    mk_prim2(PR_PUT, "put", MUST_BE_N(3));
    mk_prim2(PR_GET, "get", MUST_BE_N(2));
    mk_prim2(PR_VECTORP, "vector?", MUST_BE_N(1));
    mk_prim2(PR_MKVECTOR, "make-vector", AT_LEAST_N(1));
    mk_prim2(PR_VECLEN, "vector-length", MUST_BE_N(1));
    mk_prim2(PR_VECREF, "vector-ref", MUST_BE_N(2));
    mk_prim2(PR_VECSET, "vector-set!", MUST_BE_N(3));
    mk_prim2(PR_VECLIST, "vector->list", MUST_BE_N(1));
    mk_prim2(PR_LISTVEC, "list->vector", MUST_BE_N(1));
    mk_prim2(PR_GC, "gc", MUST_BE_N(0));
    mk_prim2(PR_GCVERBOSE, "gc-verbose", MUST_BE_N(1));
    mk_prim2(PR_GETCLOSURE, "get-closure-code", MUST_BE_N(1));
    mk_prim2(PR_RESET, "reset", MUST_BE_N(0));
    mk_prim2(PR_LENGTH, "length", MUST_BE_N(1));
    mk_prim2(PR_ASSQ, "assq", MUST_BE_N(2));
    mk_prim2(PR_ASSV, "assv", MUST_BE_N(2));
    mk_prim2(PR_ASSOC, "assoc", MUST_BE_N(2));
    mk_prim2(PR_MEMQ, "memq", MUST_BE_N(2));
    mk_prim2(PR_MEMV, "memv", MUST_BE_N(2));
    mk_prim2(PR_MEMBER, "member", MUST_BE_N(2));
    mk_prim2(PR_MKSTR, "make-string", AT_LEAST_N(1));
    mk_prim2(PR_SLEN, "string-length", MUST_BE_N(1));
    mk_prim2(PR_STRREF, "string-ref", MUST_BE_N(2));
    mk_prim2(PR_STRSET, "string-set!", MUST_BE_N(3));
    mk_prim2(PR_STREQ, "string=?", AT_LEAST_N(2));
    mk_prim2(PR_SLESS, "string<?", AT_LEAST_N(2));
    mk_prim2(PR_STRGRE, "string>?", AT_LEAST_N(2));
    mk_prim2(PR_SLEQ, "string<=?", AT_LEAST_N(2));
    mk_prim2(PR_STRGEQ, "string>=?", AT_LEAST_N(2));
    mk_prim2(PR_SCIEQ, "string-ci=?", AT_LEAST_N(2));
    mk_prim2(PR_SCILS, "string-ci<?", AT_LEAST_N(2));
    mk_prim2(PR_SCIGRE, "string-ci>?", AT_LEAST_N(2));
    mk_prim2(PR_SCILEQ, "string-ci<=?", AT_LEAST_N(2));
    mk_prim2(PR_SCIGEQ, "string-ci>=?", AT_LEAST_N(2));
    mk_prim2(PR_SUBSTR, "substring", MUST_BE_N(3));
    mk_prim2(PR_STRAPP, "string-append", AT_LEAST_N(0));
    mk_prim2(PR_STR2NUM, "string->number", AT_LEAST_N(1));
    mk_prim2(PR_NUM2STR, "number->string", AT_LEAST_N(1));
    mk_prim2(PR_SYM2STR, "symbol->string", MUST_BE_N(1));
    mk_prim2(PR_STR2SYM, "string->symbol", MUST_BE_N(1));
    mk_prim2(PR_LISTWIDTH, "list-width", MUST_BE_N(1));
    mk_prim2(PR_CHEQ, "char=?", AT_LEAST_N(2));
    mk_prim2(PR_CHLESS, "char<?", AT_LEAST_N(2));
    mk_prim2(PR_CHGRE, "char>?", AT_LEAST_N(2));
    mk_prim2(PR_CHLEQ, "char<=?", AT_LEAST_N(2));
    mk_prim2(PR_CHGEQ, "char>=?", AT_LEAST_N(2));
    mk_prim2(PR_CCIEQ, "char-ci=?", AT_LEAST_N(2));
    mk_prim2(PR_CCILS, "char-ci<?", AT_LEAST_N(2));
    mk_prim2(PR_CCIGRE, "char-ci>?", AT_LEAST_N(2));
    mk_prim2(PR_CCILEQ, "char-ci<=?", AT_LEAST_N(2));
    mk_prim2(PR_CCIGEQ, "char-ci>=?", AT_LEAST_N(2));
    mk_prim2(PR_CHALPHA, "char-alphabetic?", MUST_BE_N(1));
    mk_prim2(PR_CHNUMP, "char-numeric?", MUST_BE_N(1));
    mk_prim2(PR_CHWHITE, "char-whitespace?", MUST_BE_N(1));
    mk_prim2(PR_CHUPPER, "char-upper-case?", MUST_BE_N(1));
    mk_prim2(PR_CHLOWER, "char-lower-case?", MUST_BE_N(1));
    mk_prim2(PR_CH2INT, "char->integer", MUST_BE_N(1));
    mk_prim2(PR_INT2CHAR, "integer->char", MUST_BE_N(1));
    mk_prim2(PR_CHUP, "char-upcase", MUST_BE_N(1));
    mk_prim2(PR_CHDOWN, "char-downcase", MUST_BE_N(1));
    mk_prim2(PR_EOFP, "eof-object?", MUST_BE_N(1));
    mk_prim2(PR_INPUTP, "input-port?", MUST_BE_N(1));
    mk_prim2(PR_CURINPUT, "current-input-port", MUST_BE_N(0));
    mk_prim2(PR_OPENINPUT, "open-input-file", MUST_BE_N(1));
    mk_prim2(PR_INPCLOSE, "close-input-port", MUST_BE_N(1));
    mk_prim2(PR_OUTPUTP, "output-port?", MUST_BE_N(1));
    mk_prim2(PR_CUROUTPUT, "current-output-port", MUST_BE_N(0));
    mk_prim2(PR_OPENOUTPUT, "open-output-file", MUST_BE_N(1));
    mk_prim2(PR_OUTCLOSE, "close-output-port", MUST_BE_N(1));
    mk_prim2(PR_EXACT, "exact?", MUST_BE_N(1));
    mk_prim2(PR_INEXACT, "inexact?", MUST_BE_N(1));
    mk_prim2(PR_GENSYM, "gensym", AT_LEAST_N(0));
    mk_prim2(PR_IORESET, "rset-io", MUST_BE_N(0));
    mk_prim2(PR_MKBOX, "box", MUST_BE_N(1));
    mk_prim2(PR_BOXP, "box?", MUST_BE_N(1));
    mk_prim2(PR_UNBOX, "unbox", MUST_BE_N(1));
    mk_prim2(PR_SETBOX, "set-box!", MUST_BE_N(2));
    mk_prim2(PR_EXIT, "exit", MUST_BE_N(0));

    /* intialization of global pointers to special symbols */   
    LAMBDA = mk_sym_from_cstr("lambda", STDCHARCASE);
    QUOTE = mk_sym_from_cstr("quote", STDCHARCASE);
    QQUOTE = mk_sym_from_cstr("quasiquote", STDCHARCASE);
    UNQUOTE = mk_sym_from_cstr("unquote", STDCHARCASE);
    UNQUOTESP = mk_sym_from_cstr("unquote-splicing", STDCHARCASE);
    APPLY_ARRAY = mk_sym_from_cstr("=>", STDCHARCASE);
    RESET_PATCH = mk_sym_from_cstr("-*-reset-patch-*-", STDCHARCASE);
    
}

/* initialization of AM-Scheme */
init_scheme()
{
    int i;
    long size;
    
#ifndef THIHK_C_DA
#ifdef USE_SIGNAL
    void LOADDS get_signal();
#endif
#endif
       
    /* initialize global variables */
    NIL = (pointer)&_NIL;
    for (i = 0 ; i < HASHTABLESIZE ; i++) oblist[i] = (pointer)&_NIL;
    T = (pointer)&_T; F = (pointer)&_F; eof_object = (pointer)&_eof_object;
#ifdef VERBOSE
    gc_verbose = 1;
#else
    gc_verbose = 0;
#endif

    _std_input._type = T_INPUT;
#ifndef THINK_C_DA
    _std_input.fp = stdin;
#else
    _std_input.ref = 0;
    _std_input.other_buffer = NULL;
    _std_input.linestatus = 0;
#endif
    _std_input.status = 1;
    _std_input.index = -1;
    _std_input.pos = 0;
    _std_input.eol = 0;
    std_input = (pointer)&_std_input;
    cur_input = std_input;
    
    _std_output._type = T_OUTPUT;
    _std_output.status = 1;
    _std_output.index = -1;
#ifndef THINK_C_DA
    _std_output.fp = stdout;
#else
    _std_output.ref = 0;
    _std_output.xpos = 0;
#endif
    std_output = (pointer)&_std_output;
    cur_output = std_output;

    Heap1 = Heap2 = NULL;
    stackbottom = NULL;
    btm_rd_stack = NULL;

    if ((Heap1 = (char *)malloc(MALLOCCAST(Heap_Size))) == NULL)
        goto INIT_FAILED;
    if ((long)Heap1 <= (long)OP_MAXNUM) {
        char *new = (char *)malloc(MALLOCCAST(Heap_Size));
        free(Heap1);
        if (new == NULL) goto INIT_FAILED;
        Heap1 = new;
    }
    if ((Heap2 = (char *)malloc(MALLOCCAST(Heap_Size))) == NULL)
        goto INIT_FAILED;
    if ((long)Heap2 <= (long)OP_MAXNUM) {
        char *new = (char *)malloc(MALLOCCAST(Heap_Size));
        free(Heap2);
        if (new == NULL) goto INIT_FAILED;
        Heap2 = new;
    }
    Current_Heap = Heap1;
    sprintf(&strbuff[0], 
           "allocated 2*%ld bytes main memory areas\n", Heap_Size);
    PrintString(std_output, &strbuff[0]);
    
    size = sizeof(pointer)*Stack_Size;    
    stackbottom = (pointer *)malloc(MALLOCCAST(size));
    if (stackbottom == NULL) goto INIT_FAILED;
    stack = stackbottom - 1;
    stacktop = stackbottom + Stack_Size;
    sprintf(&strbuff[0], 
           "allocated %d*%ld bytes stack areas\n", sizeof(pointer), Stack_Size);
    PrintString(std_output, &strbuff[0]);

    size = (long)sizeof(read_op_type)*(long)DEFAULT_READ_STACK_SIZE;    
    btm_rd_stack = (read_op_type *)malloc(MALLOCCAST(size));
    if ( btm_rd_stack == NULL) goto INIT_FAILED;
    top_rd_stack = btm_rd_stack + DEFAULT_READ_STACK_SIZE;

    /* initialize Scheme's globals */
    init_globals();

    envir = global_env;
    code = NIL;
    Is_read_complete = 0;
    
    Init_read_stack()

#ifndef THINK_C_DA
#ifdef USE_SIGNAL
    signal(SIGINT, get_signal);
#endif
#endif
    car(global_env) = cons(cons(RESET_PATCH,
                                mk_closure(cons(NIL, NIL), global_env)),
                           car(global_env));
    
    scheme_alive = 1;
    return (0);

INIT_FAILED:
    if (Heap1 != NULL) free(Heap1);
    if (Heap2 != NULL) free(Heap2);
    if (stackbottom != NULL) free(stackbottom);
    if (btm_rd_stack != NULL) free(btm_rd_stack);
    return (1);
}

#ifndef THINK_C_DA
#ifdef USE_SIGNAL
void LOADDS
get_signal(sig)
int sig;
{
    signal(SIGINT, get_signal);
    reset();
}
#endif
#endif

/* Exit AM-Scheme */
exit_scheme()
{    
    if (!scheme_alive) return;

    free(Heap1);
    free(Heap2);
    free(stackbottom);
    free(btm_rd_stack);
    reset_load();
    reset_io();
    exit_os_depend();
    scheme_alive = 0;
}
