/*
 * int.c
 *
 * This file is part of AM-Scheme.
 * AM-Scheme is coded by Atsushi Moriwaki
 *
 * This software is completely free to copy and distribute modifed version.
 * But I would appreciate it if you left my name on the code as the author.
 *
 */
 
#include "amscm.h"

int_size_adjust(x)
pointer *x;
{
    long size = integersize(*x), real_size;
    uShort *array = integerarray(*x);
    
    for (real_size = size - 1 ; real_size >= 0 ; real_size--)
        if (array[real_size] != 0) break;
    ++real_size;
    if (real_size == 0) { /* x == zero */
        ++real_size;
        integersign(*x) = 1;
    }
    UnGet_Heap((size - real_size)*sizeof(uShort), 1);
    integersize(*x) = real_size;
}

long get_longint_from_integer(x)
pointer x;
{
    long size = integersize(x), i, v;
    uShort *array = integerarray(x);
    
    for (v = 0, i = size - 1 ; i >= 0 ; i--)
        v = array[i] + v*INTEGERBASE;
    v = integersign(x) * v;
    
    return (v);
}

double int_to_real(x)
pointer x;
{
    long size = integersize(x), i;
    uShort *array = integerarray(x);
    double v;
    
    for (v = 0.0, i = size - 1 ; i >= 0 ; i--)
        v = array[i] + v*INTEGERBASE;
    v = integersign(x) * v;
    
    return (v);
}

pointer copy_integer(x)
pointer x;
{
   long size = integersize(x);
   pointer y;
   
   g_push(&x)
   y = (pointer)Get_Heap(INTEGERSIZE(size));
   g_pop(1)
   memcpy(y, x, INTEGERSIZE(size));
   return (y);
}

comp_integer(a, b)
pointer a, b;
{
    char a_sign = integersign(a),
         b_sign = integersign(b);
    int negative;

    if (a_sign > 0 && b_sign > 0) negative = 0;
    else if (a_sign > 0 && b_sign < 0) return (1);
    else if (a_sign < 0 && b_sign > 0) return (-1);
    else if (a_sign < 0 && b_sign < 0) negative = 1;

    {
        long a_size = integersize(a),
             b_size = integersize(b);
        if (a_size > b_size) {
            if (negative) return (-1);
            else return (1);
        } else if (a_size < b_size) {
            if (negative) return (1);
            else return (-1);
        } else {
            uShort *a_array = integerarray(a),
                  *b_array = integerarray(b);
            long i;
            for (i = a_size - 1 ; i >= 0 ; i--) {
                if (a_array[i] > b_array[i]) {
                    if (negative) return (-1);
                    else return (1);
                } else if (a_array[i] < b_array[i]) {
                    if (negative) return (1);
                    else return (-1);
                }
            }
            return (0);
        }
    }
}

array_add(a_abs, a_size, b_abs, b_size, new_abs)
uShort *a_abs;
long a_size;
uShort *b_abs;
long b_size;
uShort *new_abs;
{
    long carry, tmp, i;

    for (i = 0, carry = 0 ; i < b_size ; i++) {
        tmp = (long)a_abs[i] + (long)b_abs[i] + carry;
        new_abs[i] = (uShort) (tmp % INTEGERBASE);
        carry = tmp / INTEGERBASE;
    }
    for ( i = b_size; i < a_size ; i++) {
        tmp = (long)a_abs[i] + carry;
        new_abs[i] = (uShort) (tmp % INTEGERBASE);
        carry = tmp / INTEGERBASE;
    }
    new_abs[a_size] = carry;
}

pointer int_add(a, b)
pointer a, b;
{
    pointer x;
    uShort *new_array, *a_abs, *b_abs;
    pointer int_sub();
    long a_size, b_size, new_size;

    if (integersign(a) > 0 && integersign(b) < 0) {
        integersign(b) = 1;
        g_push(&b)
        x = int_sub(a, b);
        g_pop(1)
        integersign(b) = -1;
        return (x);
    } else if (integersign(a) < 0 && integersign(b) > 0) {
        integersign(a) = 1;
        g_push(&a)
        x = int_sub(b, a);
        g_pop(1)
        integersign(a) = -1;
        return (x);
    }
    
    if (integersize(a) < integersize(b)) {
            x = b; b = a; a = x;
    }
    a_size = integersize(a);
    b_size = integersize(b);
    new_size = a_size + 1; 
    g_push(&a) g_push(&b) 
    x = (pointer)Get_Heap(INTEGERSIZE(new_size));
    g_pop(2)
    new_array = integerarray(x);
    a_abs = integerarray(a);
    b_abs = integerarray(b);
    array_add(a_abs, a_size, b_abs, b_size, new_array);
    type(x) = T_INTEGER;
    integersign(x) = integersign(a);
    integersize(x) = new_size;
    int_size_adjust(&x);

    return (x);
}

array_sub(a_abs, a_size, b_abs, b_size, new_abs)
uShort *a_abs;
long a_size;
uShort *b_abs;
long b_size;
uShort *new_abs;
{
    long carry, ax, bx, i;

    for (i = 0, carry = 0 ; i < a_size; i++) {
        ax = (carry) ? ((long)a_abs[i] - 1L) : (long)a_abs[i];	/* A.K */
        bx = (i < b_size) ? (long)b_abs[i] : (long)0;
        if (ax >= bx) {
            new_abs[i] = (uShort)(ax - bx);
            carry = 0;
        } else {
            new_abs[i] = (uShort)(ax - bx + (long)INTEGERBASE);
            carry = 1;
        }
    }
}

pointer int_sub(a, b)
pointer a, b;
{
    pointer x;
    uShort *new_array, *a_abs, *b_abs;
    long a_size, b_size, new_size;
    char a_sign, ab_comp;

    if (integersign(a) > 0 && integersign(b) < 0) {
        integersign(b) = 1;
        g_push(&b)
        x = int_add(a, b);
        g_pop(1)
        integersign(b) = -1;
        return (x);
    } else if (integersign(a) < 0 && integersign(b) > 0) {
        integersign(b) = -1;
        g_push(&b)
        x = int_add(a, b);
        g_pop(1)
        integersign(b) = 1;
        return (x);
    } 

    a_sign = integersign(a);
    ab_comp = comp_integer(a, b);
    if ((a_sign > 0 && ab_comp < 0) || (a_sign < 0 && ab_comp > 0)) {
        x = int_sub(b, a);
        integersign(x) = (a_sign > 0 ) ? -1 : 1;
        return (x);
    }

    a_size = integersize(a);
    b_size = integersize(b);
    new_size = a_size; 
    g_push(&a) g_push(&b) 
    x = (pointer)Get_Heap(INTEGERSIZE(new_size));
    g_pop(2)
    new_array = integerarray(x);
    a_abs = integerarray(a);
    b_abs = integerarray(b);
    array_sub(a_abs, a_size, b_abs, b_size, new_array);
    type(x) = T_INTEGER;
    integersign(x) = integersign(a);
    integersize(x) = new_size;
    int_size_adjust(&x);

    return (x);
}

array_mul(a_abs, a_size, b_abs, b_size, new_abs, new_size)
uShort *a_abs;
long a_size;
uShort *b_abs;
long b_size;
uShort *new_abs;
long new_size;
{
    long i, j;
    unsigned long carry, tmp;

    for (i = 0 ; i < new_size ; i++) new_abs[i] = 0;
    for (j = 0 ; j < b_size; j++) {
        if (b_abs[j]) {
            for (i = 0, carry = 0; i < a_size ; i++) {
                tmp = (long)a_abs[i]*(long)b_abs[j]+(long)new_abs[i+j] + carry;
                new_abs[i+j] = (uShort)(tmp % INTEGERBASE);
                carry = tmp / INTEGERBASE;
            }
            new_abs[j+a_size] = (uShort)carry;
        }
    }
}

pointer int_mul(a, b)
pointer a, b;
{
    pointer x;
    uShort *new_array, *a_abs, *b_abs;
    long a_size, b_size, new_size;

    a_size = integersize(a);
    b_size = integersize(b);
    new_size = a_size + b_size; 
    g_push(&a) g_push(&b) 
    x = (pointer)Get_Heap(INTEGERSIZE(new_size));
    g_pop(2)
    new_array = integerarray(x);
    type(x) = T_INTEGER;
    a_abs = integerarray(a);
    b_abs = integerarray(b);
    array_mul(a_abs, a_size, b_abs, b_size, new_array, new_size);
    type(x) = T_INTEGER;
    integersign(x) = integersign(a) * integersign(b);
    integersize(x) = new_size;
    int_size_adjust(&x);

    return (x);
}

array_div(a_abs, a_size, b_abs, b_size, x, y, q, IsRemainder)
uShort *a_abs;
long a_size;
uShort *b_abs;
long b_size;
uShort *x, *y, *q;
char IsRemainder;
{
    uShort * p;
    long  T1, T2, B, Q;
    long i, j;
    uShort y1, y2;
    char f;
    
    if (b_size == 1) {
        for (i = a_size - 1, T1 = 0, B = b_abs[0] ; i >= 0 ; i--) {
            T1 = T1*INTEGERBASE + (long)a_abs[i];
            q[i] = (uShort)(T1/B);
            T1 %= B;
        }
        x[0] = T1;
        return 1;
    }
            
    /* Normalization */
    B = INTEGERBASE/((long)b_abs[b_size-1] + 1);
    for (j = 0, T1 = 0 ; j < a_size ; j++) {
        T1 += ((long)a_abs[j])*B;
        x[j] = (uShort)(T1%INTEGERBASE);
        T1 /= INTEGERBASE;
    }
    x[a_size] = (uShort)T1;
    for (j = 0, T1 = 0 ; j < b_size ; j++) {
        T1 += ((long)b_abs[j])*B;
        y[j] = (uShort)(T1%INTEGERBASE);
        T1 /= INTEGERBASE;
    }
    
    y1 = y[b_size-1]; y2 = y[b_size-2];
    for (j = a_size - b_size ; j >= 0 ; j--) {
        p = x + j;
    
        /* Calulation of Q */
        T1 = (long)p[b_size]*INTEGERBASE + (long)p[b_size-1];
        Q = T1/(long)y1;
        T2 = (long)y2*Q - (T1%(long)y1)*INTEGERBASE - (long)p[b_size-2];
        if (Q >= INTEGERBASE) {
            --Q;
            T2 -= (long)y2 + (long)y1*INTEGERBASE;
        }
        while (T2 > 0) {
            --Q;
            T2 -= (long)y2 + (long)y1*INTEGERBASE;
        }
        
        
        /* Caluculation of block qotient */ 
        for (i = 0, T1 = 0, f = 0 ; i < b_size ; i++) {
            T1 += (long)y[i]*Q;
            T2 = (f) ? (long)p[i] - T1%INTEGERBASE - 1 : 
                       (long)p[i] - T1%INTEGERBASE;
            if (T2 < 0) {
                T2 += INTEGERBASE;
                f = 1;
            }
            else f = 0;
            p[i] = (uShort)T2;
            T1 /= INTEGERBASE;
        }
        T2 = (f) ? (long)p[b_size] - T1%INTEGERBASE - 1 : 
                   (long)p[b_size] - T1%INTEGERBASE;
        if (T2 < 0) {
            Q--;
            for (i = 0, T1 = 0 ; i < b_size ; i++) {
                T1 += (long)p[i] + (long)y[i];
                p[i] = (uShort)(T1%INTEGERBASE);
                T1 /= INTEGERBASE;
            }
        }
        p[b_size] = 0;
        q[j] = (uShort)Q;
    }
    if (IsRemainder) {
        for (i = b_size - 1, T1 = 0 ; i >= 0 ; i--) {
            T1 = T1*INTEGERBASE + (long)x[i];
            x[i] = (uShort)(T1/B);
            T1 %= B;
        }
    }
    return 1;
}

pointer int_quotient(a, b)
pointer a, b;
{
    pointer x;
    uShort * new_array, * a_abs, * b_abs;
    long a_size, b_size, size;
    char a_sign, b_sign;

    if (integersize(b) == 1 && *integerarray(b) == 0)
        Error("quotient -- dived by Zero");
    a_sign = integersign(a);
    b_sign = integersign(b);
    integersign(a) = 1;
    integersign(b) = 1;
    if (comp_integer(a, b) < 0) {
       integersign(a) = a_sign;
       integersign(b) = b_sign;
       return (mki_f_number(0L));
    }
    integersign(a) = a_sign;
    integersign(b) = b_sign;

    a_size = integersize(a);
    b_size = integersize(b);
    size = (a_size - b_size + 1) + (a_size + 1) + (b_size);
    g_push(&a) g_push(&b)
    x = (pointer)Get_Heap(INTEGERSIZE(size));
    g_pop(2)
    new_array = integerarray(x);
    a_abs = integerarray(a);
    b_abs = integerarray(b);
    while (size > 0) new_array[--size] = 0;
    array_div(a_abs, a_size, b_abs, b_size,
              &new_array[(a_size - b_size + 1)],
              &new_array[(a_size - b_size + 1) + (a_size + 1)],
              &new_array[0],
              0);
    UnGet_Heap(((a_size + 1) + b_size)*sizeof(uShort), 1);
    type(x) = T_INTEGER;
    integersign(x) = integersign(a) * integersign(b);
    integersize(x) = a_size - b_size + 1;
    int_size_adjust(&x);

    return (x);
}

pointer int_remainder(a, b)
pointer a, b;
{
    pointer x;
    uShort * new_array, * a_abs, * b_abs;
    long a_size, b_size, size;
    char a_sign, b_sign;

    if (integersize(b) == 1 && *integerarray(b) == 0)
        Error("remainder -- dived by Zero");
    a_sign = integersign(a);
    b_sign = integersign(b);
    integersign(a) = 1;
    integersign(b) = 1;
    if (comp_integer(a, b) < 0) {
       integersign(a) = a_sign;
       integersign(b) = b_sign;
       return (copy_integer(a));
    }
    integersign(a) = a_sign;
    integersign(b) = b_sign;

    a_size = integersize(a);
    b_size = integersize(b);
    size = (a_size + 1) + (a_size - b_size + 1) + (b_size);
    g_push(&a) g_push(&b)
    x = (pointer)Get_Heap(INTEGERSIZE(size));
    g_pop(2)
    new_array = integerarray(x);
    a_abs = integerarray(a);
    b_abs = integerarray(b);
    while (size > 0) new_array[--size] = 0;
    array_div(a_abs, a_size, b_abs, b_size,
              &new_array[0],
              &new_array[(a_size - b_size + 1) + (a_size + 1)],
              &new_array[(a_size + 1)],
              1);
    UnGet_Heap(((a_size - b_size + 1) + (b_size))*sizeof(uShort), 1);
    type(x) = T_INTEGER;
    integersign(x) = integersign(a);
    integersize(x) = a_size + 1;
    int_size_adjust(&x);

    return (x);
}

Short_Add(num, len, a)
uShort *num;
long len;
uShort a;
{
     long c, x, i;

     x = (long)num[0] + (long)a;
     num[0] = (uShort)(x % INTEGERBASE);
     c = x / INTEGERBASE; 
     for (i = 1 ; i < len; i++) {
         x = c + (long)num[i];
         num[i] = (uShort)(x % INTEGERBASE);
         c = x / INTEGERBASE;
     }
}

Short_Mult(num, len, a)
uShort *num;
long len;
uShort a;
{
     long c = 0, x, i;

     for (i = 0 ; i < len; i++) {
         x = c + (long)num[i]*(long)a;
         num[i] = (uShort)(x % INTEGERBASE);
         c = x / INTEGERBASE;
     }
}

Short_Div(num, len, a)
uShort *num;
long len;
uShort a;
{
     long c = 0, x, i;

     for (i = len - 1 ; i >= 0 ; i--) {
         x = c*(INTEGERBASE) + (long)num[i];
         num[i] = (uShort)(x / (long)a);
         c = x % (long)a;
     }
     return ((int)c);
}

Is_Zero_Int(num, len)
register uShort *num;
register long len;
{
    register long i;
    
    for (i = 0 ; i < len ; i++)
        if (num[i] != 0) return 0;
    return 1;
}

pointer str_to_integer(o, base)
pointer o;
uShort base;
{
    long i, len, size, get_formal_size_of_int();
    uShort *array;
    uShort v, ch;
    char *num;
    pointer x;
    
    size = get_formal_size_of_int(len = strlength(o), base);
    g_push(&o)
    x = (pointer)Get_Heap(INTEGERSIZE(size));
    g_pop(1)
    integersize(x) = size;
    num = strvalue(o);
    if (*num == '-') {
        integersign(x) = -1; ++num; --len;
    } else integersign(x) = 1;
    type(x) = T_INTEGER;    
    array = integerarray(x);
    for (i = 0 ; i < size ; i++) array[i] = 0;
    for (i = 0 ; i < len ; i++) {
        ch = num[i];
        Short_Mult(array, size, base);
        if ('A' <= ch && ch <= 'F') v = ch - 'A' + 10;
        else if ('a' <= ch && ch <= 'f') v = ch - 'a' + 10;
        else v = ch - '0';
        Short_Add(array, size, v);
    }
    int_size_adjust(&x);
    return (x);
    
}

pointer int_to_string(o, base)
pointer o;
uShort base;
{
    long size, len, i, bottom;
    uShort *array;
    char *new;
    uShort v;
    pointer x;

    g_push(&o)    
    size = integersize(o);
    array = (uShort *)Get_Heap(size * sizeof(uShort));
    memcpy(array, integerarray(o), size * sizeof(uShort));
    len = 0;
    do {
        ++len;
        Short_Div(array, size, base);
    } while (!Is_Zero_Int(array, size));
    if (integersign(o) < 0) ++len;
    UnGet_Heap(size * sizeof(uShort), -1);
    x = mk_string(NULL, len, '\0');
    g_push(&x)
    array = (uShort *)Get_Heap(size * sizeof(uShort));
    memcpy(array, integerarray(o), size * sizeof(uShort));
    new = strvalue(x);
    if (integersign(o) < 0) {
        new[0] = '-'; bottom = 1;
    } else bottom = 0;
    for ( i = len - 1; i >= bottom ; i--) {
        v = Short_Div(array, size, base);
        new[i] = (v >= 10) ? v - 10 + 'A' : v + '0';
    }
    UnGet_Heap(size * sizeof(uShort), -1);
    g_pop(2)
    return (x);
}
