/*
 * mac_da_os.c
 *
 * This file is part of AM-Scheme.
 * AM-Scheme is coded by Atsushi Moriwaki
 *
 * This software is completely free to copy and distribute modifed version.
 * But I would appreciate it if you left my name on the code as the author.
 *
 */
 
#include <DeviceMgr.h>
#include <WindowMgr.h>
#include <MenuMgr.h>
#include <EventMgr.h>
#include <FileMgr.h>
#include <ToolboxUtil.h>
#include <StdFilePkg.h>
#include <pascal.h>
#include <HFS.h>
#include <SetUpA4.h>
#include "amscm.h"

#define MENUID        0
#define ABOUTDIALOGID 0
#define MEMDIALOGID   1
#define MEMPARAMID    0
#define STACKPARAMID  1
#define GC_CURID      0
#define SETDEFID      2

#define TIMEON      30  /* cursor on time */
#define TIMEOFF     15  /* cursor off time */
#define LPAREN_CHECKTIME 10 /* sleep time for checking corresponding left paren */
#define TextMargin  4   /* text margin */

int             already_open = 0;   /*  1 -> DA is already open     */
DCtlPtr         dce;                /*  device control entry        */
MenuHandle      menu = 0L;          /*  handle for our menu         */
int             menuID;             /*  our menu ID                 */
WindowPtr       wp = 0L;            /*  our window pointer          */
FontInfo        fInfo;              /*  font imformation of our window */
Cursor			gc_cur;             /*  a copy of the gc cursor  */

/* cursor variables */
long cursortime;
int cursorstate;
int posx, posy;
int xinc,yinc;
int tmargin,lmargin;

char Screen[SCREENHEIGHT][SCREENWIDTH];
int scrh,scrw;
int CurScreenTop;

int do_paste = 0;
long paste_length;
Handle paste_h = NULL;

int verbose_eval = 0;

#define LPMARKSIZE 256
struct { int x; int y; } lpmark_stack[LPMARKSIZE];
int lpmark_top = -1;

main(p,d,n)
cntrlParam *p;
DCtlPtr     d;
int         n;
{
    int ret;
    
    if (d->dCtlStorage == 0) {
        if (n == 0) {
            SysBeep(3);
            CloseDriver(d->dCtlRefNum);
        }
        return(0);
    }
    
    dce = d;
    dce->dCtlFlags &= ~dCtlEnable;

    if (ret=setjmp(reset_jmp)) {
        switch (ret) {
           case 1:
               goto END;
           default:
               CloseDriver(dce->dCtlRefNum);
               return (0);
        }
    }

    switch (n) {
        case 0:
            doOpen();                         
            break;  
        case 2:
            doControl(p->csCode, p->csParam); 
            break;  
        case 4:
            doClose();
            break;  
        default: ;
    }

END:    
    dce->dCtlFlags |= dCtlEnable;
    return(0);
}

doOpen()
{
    GrafPtr     currentPort;
    Handle      h;
    long x;
    
    dce->dCtlFlags |= dNeedLock|dNeedGoodBye|dNeedTime;
    if (wp) 
        SelectWindow(wp);
    if (already_open) {
        dce->dCtlMenu = menuID;
        return;
    }

    Get_Gc_Cursor();
    already_open = 1;
    
    Create_Menu();
    Create_Window();
    PrintString(std_output, Banner);
    h = GetResource('#MEM', OwnedResourceID( MEMPARAMID ));
    Heap_Size = **((long **)h);
    ReleaseResource(h);
    h = GetResource('#MEM', OwnedResourceID( STACKPARAMID ));
    Stack_Size = **((long **)h);
    ReleaseResource(h);
    if (init_scheme()) {
        CloseDriver(dce->dCtlRefNum);
        return;
    }
    LoadInit();
    return;
}

doClose()
{
    exit_scheme();
    DeleteMenu(dce->dCtlMenu);
    DrawMenuBar();
    dce->dCtlMenu = 0;
    DisposeMenu(menu);
    Destroy_Window();
}

static
Create_Menu()
{
    dce->dCtlMenu = menuID = OwnedResourceID( MENUID );
    menu = GetMenu(menuID);
    (**menu).menuID = menuID;
    InsertMenu(menu = GetMenu(menuID), 0);
    DrawMenuBar();
}

Destroy_Window()
{
    if (wp) {
        DisposeWindow( wp );
        wp = 0L;
        dce->dCtlWindow = 0;    /* courtesy of TLL */
    }
}

static
Update_Menu_Bar()
{
    register WindowPeek wPtr = (WindowPeek) FrontWindow ();

    if (wPtr) {     /* there should always be a window, but we'll check. */
        if ( wPtr->windowKind == dce->dCtlRefNum ) {
            if (!GetMHandle (dce->dCtlMenu)) {
                InsertMenu (menu, 0);
                DrawMenuBar ();
            }
        }
        else if (GetMHandle (dce->dCtlMenu)) {
                DeleteMenu (dce->dCtlMenu);
                DrawMenuBar ();
        }
    }
}

LoadInit()
{
    Handle init;
    long length;
    
    PrintString(std_output, ";loading 'INIT' resource\n");    
    init = GetResource('INIT', OwnedResourceID(0));
    if ((length = GetHandleSize(init)) > 0) {
        HLock(init);
        inputeol(std_input) = length;
        inputpos(std_input) = 0;
        inputotherbuff(std_input) = (char *)*init;
        do {
            Read_From_Input(std_input);
            while (!eval_status) Eval_Cycle();
        } while (inputpos(std_input) < inputeol(std_input));
        clear_inputline(std_input)
        inputotherbuff(std_input) = NULL;
        eval_status = EVAL_ST_NORMAL;
        HUnlock(init);
    }
    ReleaseResource(init);
    verbose_eval = 1;
}

Get_Line(l)
pointer l;
{
    char ch, *p = inputbuffer(l);
    long count, curpos, eofpos;
    int err;

    clear_inputline(l)
    GetFPos(inputfile(l), &curpos);
    GetEOF(inputfile(l), &eofpos);
    if (curpos >= eofpos) return (EOF);   
    while(curpos < eofpos) {
        count = 1L;
        err = FSRead(inputfile(l), &count, &ch);
        ++curpos;
        if (err != noErr || count < 1) {
            break;
        } else if (ch == '\r') {
            *p++ = '\r'; inputeol(l) += 1;
            break;
         } else {
             *p++ = ch; inputeol(l) += 1;
         }
     }
     inputlinest(l) = 1;
     return (1);
}

LoadFromFileName(name)
char *name;
{
    int ref, err;
    pointer x;
    
    CtoPstr(name);
    err = OpenReadFile(name, &ref);
    PtoCstr(name);
    if (err != noErr) {
        Error("load -- unable to open %s", name);
    }
    if (load_files >= MAXFILE) {
        FSClose(ref);
        Error("load -- too many open files");
    }
    ld_file_list[load_files++] = mk_input(ref, 1);
    sprintf(&strbuff[0], ";loading %s\n", name);
    PrintString(std_output, &strbuff[0]);
}

OpenReadFile(fileName, refNum )
char    *fileName;
int     *refNum;
{
    HParamBlockRec pb;
    
    pb.ioParam.ioCompletion = 0L;
    pb.ioParam.ioNamePtr = (StringPtr) fileName;
    pb.ioParam.ioVRefNum = -SFSaveDisk;
    pb.ioParam.ioMisc = 0L;
    pb.ioParam.ioVersNum = 0;
    pb.ioParam.ioPermssn = fsRdPerm;
    pb.fileParam.ioDirID = CurDirStore;
    PBHOpen( &pb, 0 );
    *refNum = pb.ioParam.ioRefNum;
    return pb.ioParam.ioResult;
}

OpenWriteFile(fileName, refNum )
char    *fileName;
int     *refNum;
{
    HParamBlockRec pb;
    int err;

    pb.ioParam.ioCompletion = 0L;
    pb.ioParam.ioNamePtr = (StringPtr) fileName;
    pb.ioParam.ioVRefNum = -SFSaveDisk;
    pb.fileParam.ioDirID = CurDirStore;
    PBHCreate( &pb, 0 );
    if (pb.ioParam.ioResult == noErr) {
        pb.ioParam.ioCompletion = 0L;
        pb.ioParam.ioNamePtr = (StringPtr) fileName;
        pb.ioParam.ioVRefNum = -SFSaveDisk;
        pb.fileParam.ioFDirIndex = 0;
        pb.fileParam.ioDirID = CurDirStore;
        PBHGetFInfo( &pb, 0 );
        pb.ioParam.ioCompletion = 0L;
        pb.ioParam.ioNamePtr = (StringPtr) fileName;
        pb.ioParam.ioVRefNum = -SFSaveDisk;
        pb.fileParam.ioDirID = CurDirStore;
        pb.fileParam.ioFlFndrInfo.fdType = 'TEXT';
        pb.fileParam.ioFlFndrInfo.fdCreator = 'MPS ';
        PBHSetFInfo( &pb, 0 );
    } else if (pb.ioParam.ioResult != dupFNErr)
        return (pb.ioParam.ioResult);
    
    pb.ioParam.ioCompletion = 0L;
    pb.ioParam.ioNamePtr = (StringPtr) fileName;
    pb.ioParam.ioVRefNum = -SFSaveDisk;
    pb.ioParam.ioMisc = 0L;
    pb.ioParam.ioVersNum = 0;
    pb.ioParam.ioPermssn = fsWrPerm;
    pb.fileParam.ioDirID = CurDirStore;
    PBHOpen( &pb, 0 );
    if (pb.ioParam.ioResult == noErr) 
        SetEOF(pb.ioParam.ioRefNum, 0L);
    *refNum = pb.ioParam.ioRefNum;
    return pb.ioParam.ioResult;
}
    
doMenu(i)
int i;
{
    switch (i) {
        case 1:  /* about */
        {
            DialogPtr dp = GetNewDialog( OwnedResourceID(ABOUTDIALOGID), 
                                         0L, -1L );
            EventRecord e;
    
            DrawDialog(dp);
            while (!GetNextEvent(mDownMask, &e)) ;
            DisposDialog(dp);
            break;
        }
        case 3:  /* Load */
        {
            SFReply reply;
            Point   where = { 100, 104 };
            SFTypeList types = {'TEXT'};
            int ref, err;
            char *name;
            
            SFGetFile(where, "\pload file:", 0L, 1, types, 0L, &reply);
            doUpdate();
            if (!reply.good) break;
            name = PtoCstr((char *)reply.fName);
            LoadFromFileName(name);
            break;
        }
        case 4:  /* Reset */
            reset();
            break;
        case 5:  /* Reset IO */
            reset_io();
            break;
        case 6:  /* Set Memory */
       {
            DialogPtr dp = GetNewDialog( OwnedResourceID(MEMDIALOGID), 
                                         0L, -1L );
            Rect box;
            Handle mem, stack, ok, cancel, resmem, resstack;
            int Type, done = 0, select;
            long h_size, s_size;

            DrawDialog(dp);

            GetDItem(dp, 1, &Type, &mem, &box);
            GetDItem(dp, 2, &Type, &stack, &box);
            GetDItem(dp, 3, &Type, &ok, &box);
            GetDItem(dp, 4, &Type, &cancel, &box);
            
            resmem = GetResource('#MEM', OwnedResourceID( MEMPARAMID ));
            resstack = GetResource('#MEM', OwnedResourceID( STACKPARAMID ));
            h_size = **((long **)resmem);
            sprintf(&strbuff[0], "%ld", h_size);
            CtoPstr(&strbuff[0]);
            SetIText(mem, &strbuff[0]);
            s_size = **((long **)resstack);
            sprintf(&strbuff[0], "%ld", s_size);
            CtoPstr(&strbuff[0]);
            SetIText(stack, &strbuff[0]);
            
            while (!done) {
                ModalDialog(0L, &select);
                switch (select) {
                case 3:  /* ok */
                    GetIText(mem, &strbuff[0]);
                    PtoCstr(&strbuff[0]);
                    h_size = atol(&strbuff[0]);
                    **((long **)resmem) = h_size;
                    ChangedResource(resmem);
                    WriteResource(resmem);
                    GetIText(stack, &strbuff[0]);
                    PtoCstr(&strbuff[0]);
                    s_size = atol(&strbuff[0]);
                    **((long **)resstack) = s_size;
                    ChangedResource(resstack);
                    WriteResource(resstack);
                    done = 1;
                    break;
                case 4:  /* cancel */
                    done = 1;
                    break;
                }
            }
            ReleaseResource(resmem);
            ReleaseResource(resstack);
            DisposDialog(dp);
            break;
        }
        case 7:  /* Set Current Directory */
            SetDefault();
            break;
        case 8: /* Dump Init File */
        {
            SFReply reply;
            Point   where = { 100, 104 };
            SFTypeList types = {'TEXT'};
            int ref, err;
            char *name;
            Handle init, old;
            long count;
            
            SFGetFile(where, "\pinit file:", 0L, 1, types, 0L, &reply);
            doUpdate();
            if (!reply.good) break;
            err = OpenReadFile(reply.fName, &ref);
            name = PtoCstr((char *)reply.fName);
            if (err != noErr)
                Error("Dump Init File -- unable to open %s", name);
            GetEOF(ref, &count);
            if ((init = NewHandle(count)) == NULL) {
                FSClose(ref);
                Error("Dump Init File -- unable to allocate memory for dump init");
            }
            HLock(init);
            err = FSRead(ref, &count, *init);
            if (err != noErr) {
                FSClose(ref);
                DisposHandle(init);
                Error("Dump Init File -- reading error");
            }
            HUnlock(init);
            old = GetResource('INIT', OwnedResourceID(0));
            RmveResource(old);
            DisposHandle(old);
            AddResource(init, 'INIT', OwnedResourceID(0), "\p");
            ChangedResource(init);
            WriteResource(init);
            ReleaseResource(init);              
            FSClose(ref);
            break;
        }
        case 10:  /* Quit */
            CloseDriver(dce->dCtlRefNum);
            return;
        default:
            break;
    }
    HiliteMenu(0);
}

MaintainMenu()
{
    switch (eval_status) {
    case EVAL_ST_NORMAL:
        if (load_files > 0) {
            EnableItem( menu, 1 );  /* about */
            DisableItem( menu, 3 ); /* load */
            EnableItem( menu, 4 );  /* reset */
            DisableItem( menu, 5 ); /* reset io */
            DisableItem( menu, 6 ); /* set mem */
            DisableItem( menu, 7 ); /* set dir */
            DisableItem( menu, 8 ); /* dump init */
            EnableItem( menu, 10 );  /* quit */
        } else {
            EnableItem( menu, 1 ); /* about */
            EnableItem( menu, 3 ); /* load */
            EnableItem( menu, 4 ); /* reset */
            EnableItem( menu, 5 ); /* reset io */
            EnableItem( menu, 6 ); /* set mem */
            EnableItem( menu, 7 ); /* set dir */
            EnableItem( menu, 8 ); /* dump init */
            EnableItem( menu, 10 ); /* quit */
        }
		break;
    case EVAL_ST_PR_READ:
    case EVAL_ST_EXEC:
    default:
        EnableItem( menu, 1 );  /* about */
        DisableItem( menu, 3 ); /* load */
        EnableItem( menu, 4 );  /* reset */
        DisableItem( menu, 5 ); /* reset io */
        DisableItem( menu, 6 ); /* set mem */
        DisableItem( menu, 7 ); /* set dir */
        DisableItem( menu, 8 ); /* dump init */
        EnableItem( menu, 10 );  /* quit */
		break;
    }
}

#define WINDOWWIDTH     444
#define WINDOWHEIGHT    276
#define WINDOWTOP       40
#define WINDOWLEFT      10
#define WINDOWBOTTOM    WINDOWTOP + WINDOWHEIGHT
#define WINDOWRIGHT     WINDOWLEFT + WINDOWWIDTH
#define SBARWIDTH       16
#define SBARLEFT        WINDOWWIDTH - SBARWIDTH + 1
#define SBARRIGHT       WINDOWWIDTH + 1
#define SBARBOTTOM      WINDOWHEIGHT - SBARWIDTH + 2

static
Create_Window()
{
    GrafPtr savePort;
    static  Rect    bounds = {  WINDOWTOP, WINDOWLEFT, 
                                WINDOWBOTTOM, WINDOWRIGHT },
                    scrollBounds = { -1, SBARLEFT,
                                     SBARBOTTOM, SBARRIGHT };
    int i;

    GetPort(&savePort);
    wp = NewWindow( 0L, &bounds, "\pAM-Scheme", FALSE, 8, 0L, TRUE, 0L );
    
    ((WindowPeek)wp)->windowKind = dce->dCtlRefNum;
    dce->dCtlWindow = wp;
    
    SetPort(wp);
    TextFont(monaco);
    TextSize(9);
    GetFontInfo(&fInfo);
    SetPort(savePort);
    
    ShowWindow(wp);
    SelectWindow(wp);
    
    tmargin = TextMargin + fInfo.ascent;
    lmargin = TextMargin;
    xinc = fInfo.widMax;
    yinc = fInfo.ascent + fInfo.descent + fInfo.leading;
    posx = posy = 0;
    scrh = (WINDOWHEIGHT - 16 - lmargin)/yinc - 1;
    scrw = (WINDOWWIDTH - 16 - lmargin)/xinc - 1;
    CurScreenTop = 0;
    for (i = 0 ; i < SCREENHEIGHT ; i++)
        Screen[i][0] = '\0';
    cursortime = TickCount();
    cursorstate = 0;
}

Eval_Task()
{
    if (!eval_status) Eval_Cycle();
    else {
        MaintainMenu();
        switch (eval_status) {
        case EVAL_ST_NORMAL:
            if (load_files) {
                Read_From_Input(ld_file_list[load_files-1]);
                if (r_value == eof_object) {
                    FSClose(inputfile(ld_file_list[load_files-1]));
                    --load_files;
                    r_value = NIL;
                }
            } else if (isactiveline(std_input))
                Read_From_Input(std_input);
            else {
                if (do_paste > 0) {
                    inputeol(std_input) = paste_length;
                    inputpos(std_input) = 0;
                    inputotherbuff(std_input) = (char *)*paste_h;
                    inputlinest(std_input) = 1;
                    do_paste = -1;
                } else if (do_paste < 0) {
                    HUnlock(paste_h);
                    DisposHandle(paste_h);
                    paste_h = NULL;
                    inputotherbuff(std_input) = NULL;
                    clear_inputline(std_input)
                    do_paste = 0;
                } else CursorUpdate();
            }
            break;
        case EVAL_ST_PR_READ:
            if (isactiveline(std_input))
                Read_From_Input(std_input);
            else
                CursorUpdate();
            break;
        case EVAL_ST_PR_READCHAR:
            if (isactiveline(std_input)) {
                    eval_status = EVAL_ST_EXEC;
                    eval_operator = OP_RCHR1;
            } else 
                CursorUpdate();
            break;
        case EVAL_ST_PR_PEEKCHAR:
            if (isactiveline(std_input)) {
                    eval_status = EVAL_ST_EXEC;
                    eval_operator = OP_PKCH1;
            } else 
                CursorUpdate();
            break;
        default:
            CursorUpdate();
            break;
        }
    }
}

doControl(code, parm)
int code;
int *parm;
{
    Eval_Task();
    switch (code) {
        case accMenu:  
            doMenu(parm[1]);
            break;
        case accEvent: 
            doEvent(*((EventRecord **)parm));
            break;
        case accPaste:
            doPaste();
            break;
        case goodBye:
            reset_load();
            reset_io();
            break;
        default:
            break;
    }
    if (code != accCursor) Update_Menu_Bar();
}

doEvent(e)
EventRecord *e;
{
    switch (e->what) {
        case updateEvt: 
            doUpdate(); 
            break;
        case mouseDown: 
            doMouse(e->where);
            break;
        case activateEvt: 
            doActivate((Boolean)(e->modifiers & activeFlag));
            break;
        case keyDown: 
        case autoKey:
            if ((e->modifiers & cmdKey) == 0)
                doKeyEvent((char)(e->message & charCodeMask));
            else {
                switch ((e->message & charCodeMask)) {
                case 'v' : doPaste();
                           break;
                default:   break;
                }
            }
            break;
        default:
            break;
    }
}

doPaste()
{
    paste_h = NewHandle(0L);
    if ((paste_length = GetScrap(paste_h, 'TEXT', 0L)) > 0) {
        HLock(paste_h);
        do_paste = 1;
        if (cursorstate) {
                GrafPtr savePort;
                GetPort( &savePort );
                SetPort( wp );
                DrawWhite(posx, posy);
                SetPort( savePort );
         }
    } else {
        DisposHandle(paste_h);
        paste_h = NULL;
    }
}

static
Boolean TestWindowRgn(wp,p,msg)
WindowPtr wp;
Point p;
int msg;
{
    ProcPtr wDef;
    
    wDef = (ProcPtr) *((WindowPeek)wp) -> windowDefProc;
    return(CallPascalL( 8, wp, (int) wHit, p, wDef )==msg);
}

doMouse(p)
Point p;
{
    if (TestWindowRgn(wp, p, wInGrow))
        doGrow(p);
    else if (TestWindowRgn(wp, p, wInZoomIn))
        doZoom(p, inZoomIn);
    else if (TestWindowRgn(wp, p, wInZoomOut))
        doZoom(p, inZoomOut);
}

doGrow(p)
Point p;
{
    GrafPtr savePort;
    long    theResult;
    Point   thePt;
    Rect    r;
    static Rect wBounds;

    GetPort( &savePort );
    SetPort( wp );
    wBounds.top = 80;
    wBounds.left = 80;
    wBounds.bottom = yinc*SCREENHEIGHT + 16;
    wBounds.right = xinc*SCREENWIDTH + 16;
    theResult = GrowWindow( wp, p, &wBounds );
    thePt.h = LoWord( theResult );
    thePt.v = HiWord( theResult );
    SizeWindow( wp, thePt.h, thePt.v, 1 );
    r = wp->portRect;
    scrh = (r.bottom - r.top - 16 - lmargin)/yinc - 1;
    scrw = (r.right - r.left - 16 - lmargin)/xinc - 1;
    InvalRect(&r);
    if (posx > scrw) {
        posx = 0;
        Screen[(++posy + CurScreenTop)%SCREENHEIGHT][0] = '\0';
    }
    if (posy > scrh) {
        while (posy > scrh) {
            Screen[CurScreenTop][0] = '\0';
            if (++CurScreenTop >= SCREENHEIGHT) CurScreenTop = 0;
            posy--;
        }
    }

    SetPort( savePort );
}

doZoom(p,msg)
Point p;
int msg;
{
    GrafPtr savePort;
    Rect r;
    
    GetPort(&savePort);
    SetPort(wp);
    if (TrackBox(wp,p,msg)) {
        EraseRect(&wp->portRect);
        ZoomWindow(wp,msg,false);
    }
    r = wp->portRect;
    if ((scrh = (r.bottom - r.top - 16 - lmargin)/yinc - 1)
            >= SCREENHEIGHT) {
        scrh = SCREENHEIGHT - 1;
    }
    if ((scrw = (r.right - r.left - 16 - lmargin)/xinc - 1)
            >= SCREENWIDTH) {
        scrw = SCREENWIDTH-1;
    }
    InvalRect(&r);
    if (posx > scrw) {
        posx = 0;
        Screen[(++posy + CurScreenTop)%SCREENHEIGHT][0] = '\0';
    }
    if (posy > scrh) {
        while (posy > scrh) {
            Screen[CurScreenTop][0] = '\0';
            if (++CurScreenTop >= SCREENHEIGHT) CurScreenTop = 0;
            posy--;
        }
    }
    SetPort(savePort);
}

doActivate(activate)
Boolean activate;
{
    Rect r;
    GrafPtr savePort;
    
    GetPort(&savePort);
    SetPort(wp);
    r = wp->portRect;
    r.top = r.bottom - 16;
    r.left = r.left - 16;
    InvalRect(&r);
    
    SetPort( savePort );
}

doUpdate()
{
    GrafPtr savePort;
    
    GetPort( &savePort );
    SetPort( wp );

    BeginUpdate( wp );
    EraseRect(&wp->portRect);
    DrawGrowIcon( wp );
    DrawControls( wp );
    ReDrawScreen();
    EndUpdate( wp );
    
    SetPort( savePort );
}

DrawCharPos(x, y, ch)
int x, y;
char ch;
{
    MoveTo((x * xinc) + lmargin,(y * yinc) + tmargin);
    DrawChar(ch);
}

CursorUpdate()
{   
    GrafPtr savePort;

    if (cursortime < TickCount()) {
        GetPort(&savePort);
        SetPort(wp);
        if (cursorstate) {
            DrawWhite(posx, posy);
            cursortime = TickCount() + TIMEOFF;
            cursorstate = 0;
        } else {
            DrawBlack(posx, posy);
            cursortime = TickCount() + TIMEON;
            cursorstate = 1;
        }
    SetPort(savePort);
    }
}

Sleep(t)
long t;
{
    long bt, ct, et;

    for (bt = TickCount(), et = bt + t ; ;) {
        ct = TickCount();
        if (ct >= et) break;
    }
}
        
    
DrawWhite(x, y)
int x, y;
{
    Rect r;
    r.top = (y * yinc) + tmargin - fInfo.ascent;
    r.bottom = (y * yinc) + tmargin  + fInfo.descent;
    r.left = (x * xinc) + lmargin;
    r.right = (x + 1) * xinc + lmargin;
    EraseRect(&r);
}

DrawBlack(x, y)
int x, y;
{
    Rect r;
    r.top = (y * yinc) + tmargin - fInfo.ascent;
    r.bottom = (y * yinc) + tmargin  + fInfo.descent;
    r.left = (x * xinc) + lmargin;
    r.right = (x + 1) * xinc + lmargin;
    PaintRect(&r);
}

DrawReverse(x, y)
{
    Rect r;
    r.top = (y * yinc) + tmargin - fInfo.ascent;
    r.bottom = (y * yinc) + tmargin  + fInfo.descent;
    r.left = (x * xinc) + lmargin;
    r.right = (x + 1) * xinc + lmargin;
    InvertRect(&r);
}
    
scrollup()
{
    GrafPtr savePort;
    RgnHandle updateRgn;
    Rect rect;
    int i;
    
    GetPort(&savePort);
    SetPort(wp);
    updateRgn = NewRgn();
    rect = wp->portRect;
    rect.bottom -= 16 - 1;
    rect.right -= 16 - 1;
    ScrollRect(&rect,0,-yinc,updateRgn);
    DisposeRgn(updateRgn);
    Screen[CurScreenTop][0] = '\0';
    if (++CurScreenTop >= SCREENHEIGHT) CurScreenTop = 0;
    SetPort(savePort);
    for (i = 0 ; i <= lpmark_top ; i++) 
        (lpmark_stack[i]).y -= 1;
}

doKeyEvent(ch)
char ch;
{
    switch (ch) {
    case '\010':
        if (inputeol(std_input) > 0) {
            inputeol(std_input) -= 1;
            if (posx > 0) {
                char pch = Screen[(posy + CurScreenTop)%SCREENHEIGHT][posx-1];
                if (pch == '(') --lpmark_top;
                if (pch == ')') ++lpmark_top;
            }
            PrintChar(std_output, '\010');
        }
        break;
    case '\r':
    case '\n':
        PrintChar(std_output, ch);
        inputbuffer(std_input)[inputeol(std_input)] = '\r';
        inputeol(std_input) += 1;
        inputlinest(std_input) = 1;
        break;
    default:
        if (ch == '(') lpmark_push();
        if (ch == ')') lpmark_pop();
        PrintChar(std_output, ch);
        inputbuffer(std_input)[inputeol(std_input)] = ch;
        inputeol(std_input) += 1;
        break;
    }
}

PrintChar(fp, ch)
pointer fp;
char ch;
{
    GrafPtr savePort;
    long count;

    if (fp != std_output) {
        count = 1;
        FSWrite(outputfile(fp), &count, &ch);
        return;
    }
    
    GetPort(&savePort);
    SetPort(wp);
    if (cursorstate) DrawWhite(posx, posy);
    switch (ch) {
        case '\r' :
        case '\n' :
            Screen[(posy++ + CurScreenTop)%SCREENHEIGHT][posx] = '\0';
            if (posy > scrh) {
                scrollup();
                --posy;
            }
            posx = outputxpos(fp) = 0;
            break;
        case '\010' :
            if (posx > 0) {
                DrawWhite(--posx, posy);
                outputxpos(fp) = posx;
            }
            break;
        default :
            Screen[(posy + CurScreenTop)%SCREENHEIGHT][posx] = ch;
            DrawCharPos(posx++,posy,ch);
            if (posx > scrw) {
                Screen[(posy++ + CurScreenTop)%SCREENHEIGHT][posx] = '\0';
                if (posy > scrh) {
                    scrollup();
                    --posy;
                }
                posx = 0;
            }
            outputxpos(fp) = posx;
            break;
    }
    SetPort(savePort);
}

PrintString(fp, str)
pointer fp;
char *str;
{
    char ch;

    while ((ch = *str++) != '\0') PrintChar(fp, ch);
}

PrStrWithLen(fp, str, len)
pointer fp;
char *str;
long len;
{
    char ch;
    long i;

    for (i = 0 ; i < len ; i++) PrintChar(fp, str[i]);
}

Read_From_Input(l)
pointer l;
{
    tok_type tok;
    
    g_push(&l)
    while ((tok = get_token(l)) != TOK_EOL) {
        Is_read_complete = 0;
        Read_Cycle(tok);
        if (Is_read_complete) {
            Init_read_stack()
            read_white(l);
            switch (eval_status) {
            case EVAL_ST_NORMAL:
                if (r_value == eof_object) return;
                s_save(OP_EVALEND, code)
                code = r_value;
                eval_status = EVAL_ST_EXEC;
                eval_operator = OP_EVAL;
                MaintainMenu();
                g_pop(1)
                return;
            case EVAL_ST_PR_READ:
                eval_status = EVAL_ST_EXEC;
                eval_operator = OP_READ1;
                g_pop(1)
                return;
            default:
                g_pop(1)
                return;
            }
        } 
    }
    if (Is_read_complete)
        if (l == std_input) lpmark_top = -1;
    clear_inputline(l)
    g_pop(1)
}

reset_io()
{
    int i;
    for (i = 0 ; i < input_ports ; i++) {
        inputstatus(i_port_list[i]) = 0;
        FSClose(inputfile(i_port_list[i]));
    }
    for (i = 0 ; i < output_ports ; i++) {
        outputstatus(o_port_list[i]) = 0;
        FSClose(outputfile(o_port_list[i]));
    }
    input_ports = output_ports = 0;
}

reset_load()
{
    int i;
    for (i = 0 ; i < load_files ; i++)
        FSClose(inputfile(ld_file_list[i]));
    load_files = 0;
}

reset_os_depend()
{
    if (do_paste > 0) {
        DisposHandle(paste_h);
        do_paste = 0;
    }
    lpmark_top = -1;
    eval_status = EVAL_ST_NORMAL;
}

exit_os_depend()
{
    if (do_paste > 0) DisposHandle(paste_h);
}

ReDrawScreen()
{
    int x, y;
    char *p;
    
    Screen[(posy + CurScreenTop)%SCREENHEIGHT][posx] = '\0';
    for (y = 0 ; y <= posy ; y++) {
        p = Screen[(y + CurScreenTop)%SCREENHEIGHT];
        for (x = 0 ; *p != '\0' && x <= scrw ; x++, p++)
            DrawCharPos(x,y,*p);
    }
}

Get_Gc_Cursor()
{
    CursHandle  h;
    Boolean     dispose;
	
    SetResLoad(false);
    h = GetCursor(OwnedResourceID(GC_CURID));
    dispose = (GetHandleSize((Handle)h) == 0);
    SetResLoad(true);
    BlockMove(*(h=GetCursor(OwnedResourceID(GC_CURID))), &gc_cur, sizeof(Cursor));
    if (dispose) ReleaseResource((Handle)h);
}

Gc_Cur_On()
{
    SetCursor(&gc_cur);
}

Gc_Cur_Off()
{
    InitCursor();
}

lpmark_push()
{
    if (lpmark_top < LPMARKSIZE) {
        (lpmark_stack[++lpmark_top]).x = posx;
        (lpmark_stack[lpmark_top]).y = posy;
    }
}

lpmark_pop()
{
    GrafPtr savePort;
    int x, y;
    
    if (lpmark_top < 0) return;
    
    x = (lpmark_stack[lpmark_top]).x;
    y = (lpmark_stack[lpmark_top--]).y;
    if (0 <= x && x <= scrw && 0 <= y && y <= scrh) {
        GetPort(&savePort);
        SetPort(wp);
        if (cursorstate) DrawWhite(posx, posy);
        DrawReverse(x, y);
        Sleep((long)LPAREN_CHECKTIME);
        DrawReverse(x, y);
        SetPort(savePort);
    }
}

Boolean	setButtonClicked;

pascal CheckForButton(item,theDialog)
int			item;
DialogPtr	theDialog;
{
    SetUpA4();
	if (item == 11)				/* our "Select Current Directory" button	*/
	{
		setButtonClicked=TRUE;
		item=getCancel;			/* So we can get out of the dialog			*/
	}
    RestoreA4();
	return(item);
}
	  
short SetDefault()
{
	int			err,ntypes;
	Point		loc = { 100, 104 };
	WDPBRec		aWDPBrec;
	CInfoPBRec	myCPB;
	SFReply		reply;
	SFTypeList	typeList;

	ntypes=1; 
	typeList[0]='FLDR';
	setButtonClicked=FALSE;
	SFPGetFile(loc,0,0,ntypes,typeList,CheckForButton,&reply,OwnedResourceID(SETDEFID),0L);

	if (setButtonClicked)
	{
		aWDPBrec.ioCompletion=0L;
		aWDPBrec.ioWDIndex=0;
		aWDPBrec.ioNamePtr=0L;
		aWDPBrec.ioWDProcID=0;
		aWDPBrec.ioWDDirID=CurDirStore;
		aWDPBrec.ioVRefNum=-SFSaveDisk;
	  
		if (err=PBHSetVol(&aWDPBrec,false))		/* Set default Vol and dir	*/
			return(err);

		return(0);
	 }
	 else
	 	return(1);
}

OwnedResourceID(n)
{
    return( 0xC000 + ((~(dce->dCtlRefNum))<<5) + n );
}
