/*
 * memory.c
 *
 * This file is part of AM-Scheme.
 * AM-Scheme is coded by Atsushi Moriwaki
 *
 * This software is completely free to copy and distribute modifed version.
 * But I would appreciate it if you left my name on the code as the author.
 *
 */
 
#include "amscm.h"

long Heap_Size = DEFAULT_STORE_SIZE;
char *Heap1, *Heap2;
char *Current_Heap;
long Free_Heap = 0;
char *Old_Heap;
long Stack_Size = DEFAULT_STACK_SIZE;
pointer *gc_protect_list[16];
int n_gc_protect = 0;
pointer re_allocation();

/* get new memory from Heap */
char *Get_Heap(size)	/* avoid to use "void*" since we stick with K&R */
register long size;
{
    register char *new;
    register long len;

    len = ((size + ALIGNMENT - 1)/ALIGNMENT)*ALIGNMENT;
    if (Free_Heap + len > Heap_Size) {
        gc();
        if (Free_Heap + len > Heap_Size)
            Error("run out of heap area --- rerurn to top level");
    }
    new = &Current_Heap[Free_Heap];
    Free_Heap += len;
    return (new);
}

UnGet_Heap(size, upper)
register long size;
register int upper;
{
    Free_Heap -= size;
    if (upper > 0) {
        Free_Heap = ((Free_Heap + ALIGNMENT - 1)/ALIGNMENT)*ALIGNMENT;
    } else {
        Free_Heap = (Free_Heap/ALIGNMENT)*ALIGNMENT;
    }
}

/* ========== garbage collector ========== */

/* garbage collection. parameters a, b are marked. */
gc()
{
    register pointer *q;
    register int i;
#ifndef THINK_C_DA
#ifdef USE_SIGNAL
    extern void LOADDS get_signal();
    signal(SIGINT, SIG_IGN);
#endif
#endif

#ifdef THINK_C_DA
    Gc_Cur_On();
#endif
    
    if (gc_verbose) PrintString(std_output, "gc...");
    
    Old_Heap = Current_Heap;
    Current_Heap = (Old_Heap == Heap1) ? Heap2 : Heap1;
    Free_Heap = 0;

    /* realloc system globals */
    for (i = 0 ; i < HASHTABLESIZE ; i++)
       oblist[i] = re_allocation(oblist[i]);
    global_env = re_allocation(global_env);
    LAMBDA = re_allocation(LAMBDA);
    QUOTE = re_allocation(QUOTE);
    QQUOTE = re_allocation(QQUOTE);
    UNQUOTE = re_allocation(UNQUOTE);
    UNQUOTESP = re_allocation(UNQUOTESP);
    APPLY_ARRAY = re_allocation(APPLY_ARRAY);
    RESET_PATCH = re_allocation(RESET_PATCH);

    /* realloc current registers */
    envir = re_allocation(envir);
    code = re_allocation(code);
    
    for (q = stackbottom ; q <= stack ; ++q)
        *q = re_allocation(*q);

    /* realloc variables protect variable */
    for (i = 0 ; i < n_gc_protect ; i++)
        *(gc_protect_list[i]) = re_allocation(*(gc_protect_list[i]));

    /* realloc port objects */
    for (i = 0 ; i < load_files ; i++)
        ld_file_list[i] = re_allocation(ld_file_list[i]);
    for (i = 0 ; i < input_ports ; i++)
        i_port_list[i] = re_allocation(i_port_list[i]);
    for (i = 0 ; i < output_ports ; i++)
        o_port_list[i] = re_allocation(o_port_list[i]);

    if (gc_verbose) {
       sprintf(&strbuff[0], "done. %ld bytes are free.\n", 
                            Heap_Size - Free_Heap);
       PrintString(std_output, &strbuff[0]);
    }

#ifdef THINK_C_DA
    Gc_Cur_Off();
#endif

#ifndef THINK_C_DA
#ifdef USE_SIGNAL
    signal(SIGINT, get_signal);
#endif
#endif
}

pointer re_allocation(old)
pointer old;
{
    pointer new, *oldarray, *newarray, x, y;
    register long v, i;

    if (((char *)old < Old_Heap) || (Old_Heap + Heap_Size <= (char *)old)) 
        return (old);

    if (isreallocated(old)) return (new_alloc(old));
    switch (type(old)) {
    case T_STRING:
       v = strlength(old);
       new = (pointer)Get_Heap(STRINGSIZE(v));
       memcpy(new, old, STRINGSIZE(v));
       type(old) = T_NULL;
       new_alloc(old) = new;
       return (new);

    case T_INTEGER:
       v = integersize(old);
       new = (pointer)Get_Heap(INTEGERSIZE(v));
       memcpy(new, old, INTEGERSIZE(v));
       type(old) = T_NULL;
       new_alloc(old) = new;
       return (new);
        
    case T_REAL:
       new = (pointer)Get_Heap(REALSIZE);
       type(new) = T_REAL;
       rvalue(new) = rvalue(old);
       type(old) = T_NULL;
       new_alloc(old) = new;
       return (new);

    case T_SYMBOL:
       x = symprop(old);
       v = symlen(old);
       new = (pointer)Get_Heap(SYMBOLSIZE(v));
       memcpy(new, old, SYMBOLSIZE(v));
       type(old) = T_NULL;
       new_alloc(old) = new;
       symprop(new) = re_allocation(x);
       return (new);

    case T_SYNTAX:
       x = syntaxprop(old);
       v = syntaxlen(old);
       new = (pointer)Get_Heap(SYMBOLSIZE(v));
       memcpy(new, old, SYMBOLSIZE(v));
       type(old) = T_NULL;
       new_alloc(old) = new;
       syntaxprop(new) = re_allocation(x);
       return (new);

    case T_PRIM1:
    case T_PRIM2:
       v = strlen(primname(old));
       new = (pointer)Get_Heap(PRIMITIVESIZE(v));
       memcpy(new, old, PRIMITIVESIZE(v));
       type(old) = T_NULL;
       new_alloc(old) = new;
       return (new);

    case T_PAIR:
    {
       register pointer a, d, n, o;
       new = (pointer)Get_Heap(CONSSIZE);
       type(new) = T_PAIR;
       a = car(old);
       d = cdr(old);
       type(old) = T_NULL;
       new_alloc(old) = new;
       car(new) = re_allocation(a);
    
       a = new; o = d;
       for (;;) {
           if (isreallocated(o)) {
               cdr(a) = new_alloc(o);
               return (new);
           } else if (ispair(o)) {
               n = (pointer)Get_Heap(CONSSIZE);
               cdr(a) = n;
               type(n)  = T_PAIR;
               a = car(o);
               d = cdr(o);
               type(o) = T_NULL;
               new_alloc(o) = n;
               car(n) = re_allocation(a);
               a = n; o = d;
           } else {
               cdr(a) = re_allocation(o);
               return (new);
           }
        }
    }

    case T_VECTOR:
       v = vector_len(old);
       oldarray = vector_vec(old);
       new = (pointer)Get_Heap(VECTORSIZE(v));
       newarray = vector_vec(new);
       type(new) = T_VECTOR;
       vector_len(new) = v;
       type(old) = T_NULL;
       new_alloc(old) = new;       
       for (i = 0 ; i < v ; i++)
           newarray[i] = re_allocation(oldarray[i]);
       return (new);
    
    case T_CLOSURE:
       x = closure_code(old);
       y = closure_env(old);
       new = (pointer)Get_Heap(CONSSIZE);
       type(new) = T_CLOSURE;
       type(old) = T_NULL;
       new_alloc(old) = new;
       closure_code(new) = re_allocation(x);
       closure_env(new) = re_allocation(y);
       return (new);
    
    case T_CONTINUATION:
       v = cont_len(old);
       oldarray = cont_copy(old);
       new = (pointer)Get_Heap(VECTORSIZE(v));
       newarray = cont_copy(new);
       type(new) = T_CONTINUATION;
       cont_len(new) = v;
       type(old) = T_NULL;
       new_alloc(old) = new;       
       for (i = 0 ; i < v ; i++)
           newarray[i] = re_allocation(oldarray[i]);
       return (new);
       
    case T_MACRO:
       x = closure_code(old);
       y = closure_env(old);
       new = (pointer)Get_Heap(CONSSIZE);
       type(new) = T_MACRO;
       type(old) = T_NULL;
       new_alloc(old) = new;
       closure_code(new) = re_allocation(x);
       closure_env(new) = re_allocation(y);
       return (new);
    
    case T_INPUT:
       new = (pointer)Get_Heap(INPUTSIZE);
       memcpy(new, old, INPUTSIZE);
       type(old) = T_NULL;
       new_alloc(old) = new;
       return (new);

    case T_OUTPUT:
       new = (pointer)Get_Heap(OUTPUTSIZE);
       memcpy(new, old, OUTPUTSIZE);
       type(old) = T_NULL;
       new_alloc(old) = new;
       return (new);
    
    case T_BOX:
       x = boxcontext(old);
       new = (pointer)Get_Heap(BOXSIZE);
       type(new) = T_BOX;
       type(old) = T_NULL;
       new_alloc(old) = new;
       boxcontext(new) = re_allocation(x);
       return (new);

    defualt:
       FatalError("gc internal error: Illegal type");
    }
}

