/*
 * mk.c
 *
 * This file is part of AM-Scheme.
 * AM-Scheme is coded by Atsushi Moriwaki
 *
 * This software is completely free to copy and distribute modifed version.
 * But I would appreciate it if you left my name on the code as the author.
 *
 */
 
#include "amscm.h"

/* get new cons cell */
pointer cons(a, b)
pointer a, b;
{
    register pointer x;
    g_push(&a) g_push(&b)
    x = (pointer)Get_Heap(CONSSIZE);
    g_pop(2)
    type(x) = T_PAIR;
    car(x) = a;
    cdr(x) = b;
    return (x);
}

long get_formal_size_of_int(len, base) /* >= len*log_{INTEGERBASE}(base) */
long len;
uShort base;
{
    register long x = INTEGERBASE, n = -1;

    /* caluculate n = floor(log_{base}(INTEGERBASE)) */    
    do {
        ++n;
        x /= (long)base;
    } while (x != 0);
    return ((len + n - 1)/n);
}

pointer mki_f_string(num, len, base)
char *num;
long len;
uShort base;
{
    long i, size;
    uShort *array;
    uShort sign = 1, v, ch;
    pointer x;
    
    if (*num == '-') {
        sign = -1; ++num; --len;
    }
    size = get_formal_size_of_int(len, base);
    x = (pointer)Get_Heap(INTEGERSIZE(size));
    integersign(x) = sign;
    integersize(x) = size;
    type(x) = T_INTEGER;    
    array = integerarray(x);
    for (i = 0 ; i < size ; i++) array[i] = 0;
    for (i = 0 ; i < len ; i++) {
        ch = num[i];
        Short_Mult(array, size, base);
        if ('A' <= ch && ch <= 'F') v = ch - 'A' + 10;
        else if ('a' <= ch && ch <= 'f') v = ch - 'a' + 10;
        else v = ch - '0';
        Short_Add(array, size, v);
    }
    int_size_adjust(&x);
    return (x);
}

pointer mki_f_number(num)
long num;
{
    pointer x;
    uShort *array;
    long size;
    char sign;

    if (num < 0) {
        sign = -1;
        num = -num;
    } else {
        sign = 1;
    }
    size = ((num/INTEGERBASE) == 0) ? 1 : 2;
    x = (pointer)Get_Heap(INTEGERSIZE(size));
    integersign(x) = sign;
    array = integerarray(x);
    if (size == 1) {
        array[0] = (uShort)num;
    } else {
        array[0] = (uShort)(num%INTEGERBASE);
        array[1] = (uShort)(num/INTEGERBASE);
    }
    type(x) = T_INTEGER;
    integersize(x) = size;
    return (x);
}

/* get real atom */
pointer mk_real(r)
double r;
{
    register pointer x;
    x = (pointer)Get_Heap(REALSIZE);
    type(x) = T_REAL;
    rvalue(x) = r;
    return (x);
}

/* get new string */
pointer mk_string(str, len, ini)
char *str;
long len;
char ini;
{
    pointer x;
    register long i;
    char *newstr;

    x = (pointer)Get_Heap(STRINGSIZE(len));
    newstr = strvalue(x);
    if (str == NULL) {
        for (i = 0 ; i < len ; i++) newstr[i] = ini;
    } else {
        for (i = 0 ; i < len ; i++) newstr[i] = str[i];
    }
    newstr[len] = '\0';
    strlength(x) = len;
    type(x) = T_STRING;
    return (x);
}

/* hash function */
int hash(name, len)
register char *name;
register long len;
{
    register long x, i;
    
    for (x = 0, i = 0; i < len ; ++i) x += name[i];
    return ((int)(x % HASHTABLESIZE));
}

int str_eqv(s1, l1, s2, l2)
register char *s1;
register long l1;
register char *s2;
register long l2;
{
    register long i;
    
    if (l1 != l2) return (0);
    for (i = 0 ; i < l1 ; i++) {
        if (s1[i] != s2[i]) return (0);
    }
    return (1);
}


int UpperCase(ch)
int ch;
{
    if ('a' <= ch && ch <= 'z') return (ch - 'a' + 'A');
    else return (ch);
}

int LowerCase(ch)
int ch;
{
    if ('A' <= ch && ch <= 'Z') return (ch - 'A' + 'a');
    else return (ch);
}

/* get new symbol */
pointer mk_sym(name, len, char_case)
char *name;
long len;
int char_case;
{
    pointer x;
    register long i;
    int hashpos;

    if (char_case > 0) {
        /* change to upper characters */
        for (i = 0 ; i < len ; ++i) name[i] = UpperCase((int)name[i]);
    } else if (char_case < 0) {
        /* change to lower characters */
        for (i = 0 ; i < len ; ++i) name[i] = LowerCase((int)name[i]);
    }

    /* fisrt check oblist */
    hashpos = hash(name, len);
    for (x = oblist[hashpos]; x != NIL ; x = cdr(x)) {
        if (str_eqv(name, len, symname(car(x)), symlen(car(x))))
            break;
    }

    if (x != NIL) return (car(x));
    else {
        char *new;
        x = (pointer)Get_Heap(SYMBOLSIZE(len));
        for (i = 0, new = symname(x) ; i < len ; i++) new[i] = name[i];
        new[len] = '\0';
        symlen(x) = len;
        type(x) = T_SYMBOL;
        symprop(x) = NIL;
        g_push(&x)
        oblist[hashpos] = cons(x, oblist[hashpos]);
        g_pop(1)
        return (x);
    }
}

pointer mk_sym_from_cstr(name, char_case)
char *name;
int char_case;
{
    strcpy(&strbuff[0], name);
    return (mk_sym(&strbuff[0], (long)strlen(&strbuff[0]), char_case));
}

/* generate new symbol */
pointer gensym(head)
char *head;
{
    long symcnt = 0, len;
    register pointer x;
    int hashpos;

    do {
        /* generate a symbol, test whether it is new or not */
        if (head != NULL)
            sprintf(&strbuff[0], "%s%lu", head, symcnt++);
        else
            sprintf(&strbuff[0], "G%lu", symcnt++);
        /* check in oblist */
        len = (long)strlen(&strbuff[0]);
        hashpos = hash(&strbuff[0], len);
        for (x = oblist[hashpos]; x != NIL ; x = cdr(x)) {
            if (str_eqv(&strbuff[0], len, symname(car(x)), symlen(car(x))))
                break;
        }
    } while (x != NIL);
    x = mk_sym(&strbuff[0], len, STDCHARCASE);
    return (x);
}


int Is_Digit(ch, base)
int ch;
uShort base; /* 2, 8, 10, 16 */
{
    if (base != 16) {
        if ('0' <= ch && ch <= '0' + (int)base - 1)
            return 1;
    } else {
        if (('0' <= ch && ch <= '9') ||
            ('A' <= ch && ch <= 'F') ||
            ('a' <= ch && ch <= 'f'))
            return 1;
    }
    return 0;
}


int numtype(q, base)
char *q;
uShort base;
{
    char *p;
    int IsReal = 0;

    p = q;
    if (*p == '-' || *p == '+') ++p;
    if (!*p) return (NO_NUM_TYPE);
    while (*p && (Is_Digit(*p, base) || *p == '.' || *p == 'e')) {
        if (*p == '.') {
            if (IsReal) return (NO_NUM_TYPE);
            else IsReal = 1;
        } else if (*p == 'e' && base != 16) {
            ++p;
            if (*p == '-' || *p == '+') ++p;
            if (!*p) return (NO_NUM_TYPE);
            while (*p && isdigit(*p)) ++p;
            if (*p) return(NO_NUM_TYPE);
            else return (NUM_TYPE_REAL);
        }
        ++p;
    }
    if (*p) return(NO_NUM_TYPE);
    else if (IsReal)
        if (p == q + 1) return(NO_NUM_TYPE);
        else return (NUM_TYPE_REAL);
    else return (NUM_TYPE_INT);
}

/* make symbol or number atom from string */
pointer mk_atom(q)
char *q;
{
    switch (numtype(q, 10)) {
    case NO_NUM_TYPE:
        return (mk_sym(q, (long)strlen(q), STDCHARCASE));
    case NUM_TYPE_INT:
        return (mki_f_string(q, (long)strlen(q), 10));
    case NUM_TYPE_REAL:
        return (mk_real(atof(q)));
    }
}

/* make constant */
pointer mk_const(name)
char *name;
{
    char *body = &name[1];
    int x, i, len, inchar();
    uShort base;

    if (!strcmp(body, "t")) return (T);
    else if (!strcmp(body, "f")) return (F);
    else if (body[0] == 'b') { base = 2; goto MAKENUM; }
    else if (body[0] == 'o') { base = 8; goto MAKENUM; }
    else if (body[0] == 'd') { base = 10; goto MAKENUM; }
    else if (body[0] == 'x') { base = 16; goto MAKENUM; }
    else if (body[0] == '\\') {
        if (body[1] == '\0') /* #\ */
            return ((pointer)&char_table[' ']);
        else if (body[2] == '\0') /* #\<one char> */
            return ((pointer)&char_table[body[1]]);
        else {
            ++body;
            for (i = 0, len = strlen(body) ; i < len ; ++i)
                body[i] = LowerCase(body[i]);
            if (!strcmp(body, "space")) return((pointer)&char_table[' ']);
            else if (!strcmp(body, "newline"))
#ifndef THINK_C_DA
                return((pointer)&char_table['\n']);
#else
                return((pointer)&char_table['\r']);
#endif
            else if (!strcmp(body, "tab"))
                return((pointer)&char_table['\t']);
                            
            /* octal expression like #\012 */
            for (x = 0 ; *body ; body++) {
                if (*body >= '0' && *body <= '7')
                    x = 8*x + (int)*body - (int)'0';
                else
                    return(mk_sym(name, (long)strlen(name), STDCHARCASE));
            }
            return ((pointer)&char_table[x & CHARMASK]);
        }
    }
    return (mk_sym(name, (long)strlen(name), STDCHARCASE));

MAKENUM:
    switch (numtype(&body[1], base)) {
    case NO_NUM_TYPE:
        return (mk_sym(name, (long)strlen(name), STDCHARCASE));
    case NUM_TYPE_INT:
        return (mki_f_string(&body[1],
                                       (long)strlen(&body[1]), base));
    case NUM_TYPE_REAL:
        if (base == 10) return (mk_real(atof(&body[1])));
        else Error("radix of real type is only 10 in this system");
    }    
}

/* make vector */
pointer mk_vector(len, ini)
long len;
pointer ini;
{
    pointer x, *vec;
    register long i;
    
    g_push(&ini)
    x = (pointer)Get_Heap(VECTORSIZE(len));
    g_pop(1)
    vec = vector_vec(x);
    for (i = 0 ; i < len ; i++) vec[i] = ini;
    type(x) = T_VECTOR;
    vector_len(x) = len;

    return (x);
}   
    
/* make closure. c is code. e is environment */
pointer mk_closure(c, e)
pointer c, e;
{
    pointer x;
    
    g_push(&c) g_push(&e)
    x = (pointer)Get_Heap(CONSSIZE);
    g_pop(2)
    type(x) = T_CLOSURE;
    car(x) = c;
    cdr(x) = e;
    return (x);
}

/* make continuation. */
pointer mk_continuation(cont)
pointer *cont;
{
    pointer x, *copy, *p, *q;
    long len;

    len = ((long)cont - (long)stackbottom)/sizeof(pointer) + 1;
    x = (pointer)Get_Heap(VECTORSIZE(len));
    copy = cont_copy(x);
    for (p = stackbottom, q = copy ; p <= cont ; ++p, ++q)
        *q = *p;
    type(x) = T_CONTINUATION;
    cont_len(x) = len;

    return (x);
}

/* make input port */
pointer mk_input(f, load)
#ifndef THINK_C_DA
FILE *f;
#else
int f;
#endif
int load;
{
    pointer x = (pointer)Get_Heap(INPUTSIZE);
    type(x) = T_INPUT;
    inputfile(x) = f;
#ifdef THINK_C_DA
    inputotherbuff(x) = NULL;
#endif
    inputstatus(x) = 1;
    if (load) {
        inputindex(x) = -1;
    } else {
        inputindex(x) = input_ports;
        i_port_list[input_ports++] = x;
    }
    clear_inputline(x)
    return (x);
}

/* make output port */
pointer mk_output(f)
#ifndef THINK_C_DA
FILE *f;
#else
int f;
#endif
{
    pointer x = (pointer)Get_Heap(OUTPUTSIZE);
    type(x) = T_OUTPUT;
    outputfile(x) = f;
#ifdef THINK_C_DA
    outputxpos(x) = 0;
#endif
    outputstatus(x) = 1;
    outputindex(x) = output_ports;
    o_port_list[output_ports++] = x;
    return (x);
}

/* make box */
pointer mk_box(a)
pointer a;
{
    pointer x;
    
    g_push(&a)
    x = (pointer)Get_Heap(BOXSIZE);
    type(x) = T_BOX;
    boxcontext(x) = a;
    g_pop(1)
    return (x);
}
