/*
 * numop.c
 *
 * This file is part of AM-Scheme.
 * AM-Scheme is coded by Atsushi Moriwaki
 *
 * This software is completely free to copy and distribute modifed version.
 * But I would appreciate it if you left my name on the code as the author.
 *
 */

#include "amscm.h"

extern double int_to_real();
extern pointer copy_integer();
extern pointer int_add();
extern pointer int_sub();
extern pointer int_mul();
extern pointer int_quotient();
extern pointer int_remainder();

#define NUM_INT  0
#define NUM_REAL 1

double round(x)
double x;
{
    double y, z;
    y = floor(x);
    z = x - y;
    if (z == 0.5) {
        if (y/2 == floor(y/2)) /* y is even */
            return (y);
        else return (y + 1.0);
    } else if (z < 0.5) return (y);
    else return (y + 1.0);
}

/* numerical operators */
pointer numerical_operator(op)
register int op;
{    
    pointer ix, iy;
    double rx, ry;
    int num_type = NUM_INT;
    long i;
    int compare = 0;

    if (nargs <= 1) { /* no argument */
        switch (op) {
        case OP_ADD:
            return(mki_f_number(0L));
        case OP_MUL:
            return(mki_f_number(1L));
        default:
            Error("No argument in numerical operator");
        }
    }

    /* get one argument */
    if (isreal(args[1])) {
        num_type = NUM_REAL;
        rx = rvalue(args[1]);
    } else if (isinteger(args[1]))
        ix = args[1];
    else Error("Illegal type for numeric operator");
    
    if (nargs == 2) { /* one argument */
        switch (num_type) {
        case NUM_REAL:
            switch (op) {
            case OP_ADD:
            case OP_MUL:
            case OP_MIN:
            case OP_MAX:
                return(mk_real(rx));
            case OP_SUB:
                return(mk_real(-rx));
            case OP_DIV:
                if (rx == 0.0) Error("/ -- divided by zero");
                return(mk_real(1/rx));
            case OP_ABS:
                if (rx < 0) return(mk_real(-rx));
                else return(mk_real(rx));
            case OP_EXP:
                return(mk_real(exp(rx)));
            case OP_LOG:
                return(mk_real(log(rx)));
            case OP_SIN:
                return(mk_real(sin(rx)));
            case OP_COS:
                return(mk_real(cos(rx)));
            case OP_TAN:
                return(mk_real(tan(rx)));
            case OP_ASIN:
                return(mk_real(asin(rx)));
            case OP_ACOS:
                return(mk_real(acos(rx)));
            case OP_ATAN:
                return(mk_real(atan(rx)));
            case OP_SQRT:
                return(mk_real(sqrt(rx)));
            case OP_ZEROP:
                if (rx == 0.0) return(T); else return(F);
            case OP_POSP:
                if (rx > 0.0) return(T); else return(F);
            case OP_NEGP:
                if (rx < 0.0) return(T); else return(F);
            case OP_NEQ:
            case OP_LESS:
            case OP_GRE:
            case OP_LEQ:
            case OP_GEQ:
                return (T);
            case OP_EVEN:
            case OP_ODD:
                Error("require integer as argument");
            case OP_FLOOR:
                return(mk_real(floor(rx)));
            case OP_CEIL:
                return(mk_real(ceil(rx)));
            case OP_TRUN:
                if (rx >= 0) return(mk_real(floor(rx)));
                else return(mk_real(ceil(rx)));
            case OP_ROUND:
                return(mk_real(round(rx)));
            default:
                Error("Few arguments in numerical operator");
            }
        case NUM_INT:
            switch (op) {
            case OP_ADD:
            case OP_MUL:
            case OP_MIN:
            case OP_MAX:
            case OP_FLOOR:
            case OP_CEIL:
            case OP_TRUN:
            case OP_ROUND:
                ix = copy_integer(ix);
                return(ix);
            case OP_SUB:
                ix = copy_integer(ix);
                if (integersize(ix) != 1 || (integerarray(ix))[0] != 0) {
                    integersign(ix) *= -1;
                }
                return(ix);
            case OP_DIV:
                if (integersize(ix) == 1 && (integerarray(ix))[0] == 0)
                    Error("/ -- divided by zero");
                return(mk_real(1/int_to_real(ix)));
            case OP_ABS:
                ix = copy_integer(ix);
                integersign(ix) = 1;
                return (ix);
            case OP_ODD:
                if ((integerarray(ix))[0] % 2) return(T);
                else return (F);
            case OP_EVEN:
                if ((integerarray(ix))[0] % 2) return(F);
                else return (T);
            case OP_EXP:
                return(mk_real(exp(int_to_real(ix))));
            case OP_LOG:
                return(mk_real(log(int_to_real(ix))));
            case OP_SIN:
                return(mk_real(sin(int_to_real(ix))));
            case OP_COS:
                return(mk_real(cos(int_to_real(ix))));
            case OP_TAN:
                return(mk_real(tan(int_to_real(ix))));
            case OP_ASIN:
                return(mk_real(asin(int_to_real(ix))));
            case OP_ACOS:
                return(mk_real(acos(int_to_real(ix))));
            case OP_ATAN:
                return(mk_real(atan(int_to_real(ix))));
            case OP_SQRT:
                return(mk_real(sqrt(int_to_real(ix))));
            case OP_ZEROP:
                if (integersize(ix) == 1 && (integerarray(ix))[0] == 0)
                    return(T);
                else return(F);
            case OP_POSP:
                if (integersign(ix) > 0 &&
                    (integersize(ix) != 1 || (integerarray(ix))[0] != 0)) 
                    return(T);
                else return(F);
            case OP_NEGP:
                if (integersign(ix) < 0)
                    return(T);
                else return(F);
            case OP_NEQ:
            case OP_LESS:
            case OP_GRE:
            case OP_LEQ:
            case OP_GEQ:
                return (T);
            default:
                Error("Few arguments in numerical operator");
            }
        }
    }
    
    if (op==OP_NEQ||op==OP_LESS||op==OP_GRE||op==OP_LEQ||op==OP_GEQ)
        compare = 1;
    for (i = 2; i < nargs ; i++) { /* more arguments */
        switch (num_type) {
        case NUM_INT:
            if (isreal(args[i])) {
                num_type = NUM_REAL;
                rx = int_to_real(ix);
                ry = rvalue(args[i]);
            } else if (isinteger(args[i]))
                iy = args[i];
            else Error("Illegal type for numerical operator");
            break;
        case NUM_REAL:
            if (isreal(args[i])) {
                ry = rvalue(args[i]);
            } else if (isinteger(args[i]))
                ry = int_to_real(args[i]);
            else Error("Illegal type for numerical operator");
            break;
        }
        
        switch (num_type) {
        case NUM_REAL:
            switch (op) {
            case OP_ADD: rx += ry; break;
            case OP_SUB: rx -= ry; break;
            case OP_MUL: rx *= ry; break;
            case OP_DIV:
                if (ry == 0.0) Error("/ -- divided by zero");
                rx /= ry; break;
            case OP_QUO:
            case OP_REM:
                Error("Every argument must be integer");
            case OP_MIN:
                if (rx > ry) rx = ry;
                break;
            case OP_MAX:
                if (rx < ry) rx = ry;
                break;
            case OP_NEQ: 
                if (rx != ry) return(F);
                break;
            case OP_LESS:
                if (rx < ry) rx = ry; else return(F);
                break;
            case OP_GRE:
                if (rx > ry) rx = ry; else return(F);
                break;
            case OP_LEQ:
                if (rx <= ry) rx = ry; else return(F);
                break;
            case OP_GEQ:
                if (rx >= ry) rx = ry; else return(F);
                break;
            case OP_EXPT:
                return(mk_real(pow(rx, ry)));
            default:
                Error("Maybe many arguments in numerical operator");
            }
            break;
        case NUM_INT:
            switch (op) {
            case OP_ADD: ix = int_add(ix, iy); break;
            case OP_SUB: ix = int_sub(ix, iy); break;
            case OP_MUL: ix = int_mul(ix, iy); break;
            case OP_QUO:
                return(int_quotient(ix, iy));
            case OP_REM:
                return(int_remainder(ix, iy));
            case OP_MOD:
                if ((integersign(ix) * integersign(iy)) > 0) {
                    return(int_remainder(ix, iy));
                } else {
                    ix = int_remainder(ix, iy);
                    return (int_add(ix , iy));
                } 
            case OP_DIV:
                num_type = NUM_REAL;
                rx = int_to_real(ix) / int_to_real(iy);
                break;
            case OP_MIN:
                if (comp_integer(ix, iy) > 0) ix = iy;
                break;
            case OP_MAX:
                if (comp_integer(ix, iy) < 0) ix = iy;
                break;
            case OP_NEQ: 
                if (comp_integer(ix, iy) != 0) return(F);
                break;
            case OP_LESS:
                if (comp_integer(ix, iy) < 0) ix = iy; else return(F);
                break;
            case OP_GRE:
                if (comp_integer(ix, iy) > 0) ix = iy; else return(F);
                break;
            case OP_LEQ:
                if (comp_integer(ix, iy) <= 0) ix = iy; else return(F);
                break;
            case OP_GEQ:
                if (comp_integer(ix, iy) >= 0) ix = iy; else return(F);
                break;
            case OP_EXPT:
                return(mk_real(pow(int_to_real(ix),
                                   int_to_real(iy))));
            default:
                Error("Maybe many arguments in numerical operator");
            }
            break;
        }
    }
    if (compare) return(T);
    switch (num_type) {
    case NUM_REAL:
        return(mk_real(rx));
    case NUM_INT:
        return(ix);
    }
}
