/*
 * prim.c
 *
 * This file is part of AM-Scheme.
 * AM-Scheme is coded by Atsushi Moriwaki
 *
 * This software is completely free to copy and distribute modifed version.
 * But I would appreciate it if you left my name on the code as the author.
 *
 */

#include "amscm.h"

extern int UpperCase();
extern int LowerCase();


long list_length(l)
register pointer l;
{
    register long i = 0;
    register pointer a = l;
    do {
        if (!ispair(l)) return ((l == NIL) ? i : (long)(-1));
        l = cdr(l);
        i++;
        if (!ispair(l)) return ((l == NIL) ? i : (long)(-1));
        l = cdr(l);
        i++;
        a = cdr(a);
    } while (l != a);
    return ((long)(-1));
}

/* compare string lexicographic order */
int str_cmp(s1, l1, s2, l2)
register char *s1;
register long l1;
register char *s2;
register long l2;
{
    register long i;
    int inv = 1;
    
    if (l1 > l2) {
        char *s = s1; long l = l1;
        s1 = s2; l1 = l2; s2 = s; l2 = l;
        inv = -1;
    }
     /* here we assume l1 <= l2 */
    for (i = 0 ; i < l1 ; i++) {
        if (s1[i] > s2[i]) return (inv);
        else if (s1[i] < s2[i]) return (inv*(-1));
    }
    if (l1 == l2) return (0);
    else return (inv*(-1));
}

/* compare string lexicographic order (ignore upper and lower) */
int str_cmp_ci(s1, l1, s2, l2)
register char *s1;
register long l1;
register char *s2;
register long l2;
{
    register long i;
    int inv = 1;
    
    if (l1 > l2) {
        char *s = s1; long l = l1;
        s1 = s2; l1 = l2; s2 = s; l2 = l;
        inv = -1;
    }
     /* here we assume l1 <= l2 */
    for (i = 0 ; i < l1 ; i++) {
        if (LowerCase((int)s1[i]) > LowerCase((int)s2[i]))
            return (inv);
        else if (LowerCase((int)s1[i]) < LowerCase((int)s2[i]))
            return (inv*(-1));
    }
    if (l1 == l2) return (0);
    else return (inv*(-1));
}

/* equivalence of atoms */
int eqv(a, b)
register pointer a, b;
{
    switch (type(a)) {
    case T_STRING:
        if (isstring(b))
            return (str_eqv(strvalue(a), strlength(b),
                               strvalue(a), strlength(b)));
        else return (0);
    
    case T_INTEGER:
        if (isinteger(b)) return (comp_integer(a, b) == 0);
        else return (0);
    
    case T_REAL:
        if (isreal(b)) return (rvalue(a) == rvalue(b));
        else return (0);
    
    default:
        return (a == b);
   }
}

/* equivalence of list */
int equal(a, b)
register pointer a, b;
{
LOOP:
    switch (type(a)) {
    case T_PAIR:
        if (ispair(b)) {
            if (equal(car(a), car(b))) {
                a = cdr(a); b = cdr(b);
                goto LOOP;
            } else return (0);
        } else return (0);
    
    case T_VECTOR:
        if (isvector(b)) {
            long a_len = vector_len(a), b_len = vector_len(b);
            if (a_len == b_len) {
                long i;
                pointer *a_vec = vector_vec(a), *b_vec = vector_vec(b);
                for (i = 0 ; i < a_len ; i++)
                    if (!equal(a_vec[i], b_vec[i])) return (0);
                return (1);
            } else return (0);
        } else return (0);
    
    case T_BOX:
        if (isbox(b)) return (equal(boxcontext(a), boxcontext(b)));
        else return (0);
    
    default:
        return (eqv(a, b));
    }
}

pointer PR_LIST() /* list */
{
    register pointer x;
    register long i;
    
    for (i = nargs - 1, x = NIL ; i > 0 ; i--)
        x = cons(args[i], x);
    
    return (x);
}

pointer PR_APPEND() /* append */
{
    pointer g, k, ret;
    long i, count = nargs - 1;
    
    if (count == 0) return (NIL);
    for (i = 1 ; i < count ; i++)
        if (args[i] != NIL) break;
    if (i >= count) return (args[count]);
    ret = k = cons(car(args[i]), NIL);
    g = cdr(args[i]);
    g_push(&ret) g_push(&k) g_push(&g)
    while (i < count) {
        if (g != NIL) {
            if (list_length(g) < 0)
                Error("append -- arg is not list");
            while(g != NIL){
                cdr(k) = cons(car(g), NIL);
                k = cdr(k); g = cdr(g);
            }
        }
        g = args[++i];
    }
    cdr(k) = args[count];
    g_pop(3)
    return (ret);
}

pointer PR_REVERSE() /* reverse */
{
    register pointer y;
    pointer g = args[1];
    if (list_length(g) < 0) Error("reverse -- arg is not list");
    g_push(&g)
    for (y = NIL ; ispair(g) ; g = cdr(g))
        y = cons(car(g), y);
    g_pop(1)
    
    return (y);
}

pointer PR_LISTREF() /* list-ref */
{
    register long v, i;
    register pointer x;
    
    if (!isinteger(args[2]))
        Error("list-ref -- second arg is not integer");   
    v = get_longint_from_integer(args[2]);
    for (x = args[1], i = 0; i < v ; i++) {
        if (!ispair(x)) Error("list-ref -- out of range");
        x = cdr(x);
    }
    
    return(car(x));
}

pointer PR_CAR() /* car */
{    
    if (ispair(args[1])) {
        return(car(args[1]));
     } else Error("car -- unable to car for non-pair cell");
}

pointer PR_CDR() /* cdr */
{    
    if (ispair(args[1])) {
        return(cdr(args[1]));
    } else Error("cdr -- unable to cdr for non-pair cell");
}
        
pointer PR_CONS() /* cons */
{
    return (cons(args[1], args[2]));
}

pointer PR_SETCAR() /* set-car! */
{   
    if (ispair(args[1])) {
        car(args[1]) = args[2];
        return(T);
    } else Error("set-car! -- unable to set-car! for non-pair cell");
}

pointer PR_SETCDR() /* set-cdr! */
{    
    if (ispair(args[1])) {
        cdr(args[1]) = args[2];
        return(T);
    } else Error("set-cdr! -- unable to set-cdr! for non-pair cell");
}

pointer PR_ADD()  /* + */ { return (numerical_operator(OP_ADD)); }
pointer PR_SUB()  /* - */ { return (numerical_operator(OP_SUB)); }
pointer PR_MUL()  /* * */ { return (numerical_operator(OP_MUL)); }
pointer PR_DIV()  /* / */ { return (numerical_operator(OP_DIV)); }
pointer PR_REM()  /* remainder */ { return (numerical_operator(OP_REM)); }
pointer PR_MOD()  /* modulo */ { return (numerical_operator(OP_MOD)); }
pointer PR_QUO()  /* quotient */ { return (numerical_operator(OP_QUO)); }
pointer PR_NEQ()  /* = */ { return (numerical_operator(OP_NEQ)); }
pointer PR_LESS() /* < */ { return (numerical_operator(OP_LESS)); }
pointer PR_GRE()  /* > */ { return (numerical_operator(OP_GRE)); }
pointer PR_LEQ()  /* <= */ { return (numerical_operator(OP_LEQ)); }
pointer PR_GEQ()  /* >= */ { return (numerical_operator(OP_GEQ)); }
pointer PR_ABS()  /* abs */ { return (numerical_operator(OP_ABS)); }
pointer PR_MAX()  /* max */ { return (numerical_operator(OP_MAX)); }
pointer PR_MIN()  /* min */ { return (numerical_operator(OP_MIN)); }
pointer PR_EXPT() /* expt */ { return (numerical_operator(OP_EXPT)); }
pointer PR_ZEROP()/* zero? */ { return (numerical_operator(OP_ZEROP)); }
pointer PR_POSP() /* positive? */ { return (numerical_operator(OP_POSP)); }
pointer PR_NEGP() /* negative? */ { return (numerical_operator(OP_NEGP)); }
pointer PR_EVEN() /* even? */ { return (numerical_operator(OP_EVEN)); }
pointer PR_ODD()  /* odd? */ { return (numerical_operator(OP_ODD)); }
pointer PR_EXP()  /* exp */ { return (numerical_operator(OP_EXP)); }
pointer PR_LOG()  /* log */ { return (numerical_operator(OP_LOG)); }
pointer PR_SIN()  /* sin */ { return (numerical_operator(OP_SIN)); }
pointer PR_COS()  /* cos */ { return (numerical_operator(OP_COS)); }
pointer PR_TAN()  /* tan */ { return (numerical_operator(OP_TAN)); }
pointer PR_ASIN() /* asin */ { return (numerical_operator(OP_ASIN)); }
pointer PR_ACOS() /* acos */ { return (numerical_operator(OP_ASIN)); }
pointer PR_ATAN() /* atan */ { return (numerical_operator(OP_ATAN)); }
pointer PR_SQRT() /* sqrt */ { return (numerical_operator(OP_SQRT)); }
pointer PR_FLOOR() /* floor */ { return (numerical_operator(OP_FLOOR)); }
pointer PR_CEIL() /* ceiling */ { return (numerical_operator(OP_CEIL)); }
pointer PR_TRUN() /* truncate */ { return (numerical_operator(OP_TRUN)); }
pointer PR_ROUND() /* round */ { return (numerical_operator(OP_ROUND)); }

pointer PR_NULLP() /* null? */
{    
    if (args[1] == NIL) return(T);
    else return(F);
}

pointer PR_NOT() /* not */
{    
    if (isfalse(args[1])) return(T);
    else return(F);
}

pointer PR_BOOLP() /* boolean? */
{    
    if (args[1] == F || args[1] == T) return(T);
    else return(F);
}

pointer PR_SYMBOLP() /* symbol? */
{    
    if (issymbol(args[1])) return (T);
    else return(F);
}

pointer PR_CHP() /* char? */
{    
    if (ischar(args[1])) return (T);
    else return(F);
}

pointer PR_STRINGP() /*string? */
{    
    if (isstring(args[1])) return (T);
    else return(F);
}

pointer PR_INTEGERP() /* integer? */
{    
    if (isinteger(args[1])) return (T);
    else return(F);
}

pointer PR_REALP() /* real? */
{    
    if (isinteger(args[1]) || isreal(args[1])) return (T);
    else return(F);
}

pointer PR_PRIMP() /* primitive-procedure? */
{    
    if (isprim(args[1])) return (T);
    else return (F);
}

pointer PR_CLOSUREP() /* closure? */
{    
    if (isclosure(args[1])) return (T);
    else return(F);
}

pointer PR_CONTP() /* continuation? */
{    
    if (iscontinuation(args[1])) return (T);
    else return (F);
}

pointer PR_MACROP() /* macro? */
{   
    if (ismacro(args[1])) return (T);
    else return (F);
}

pointer PR_PAIRP() /* pair? */
{    
    if (ispair(args[1])) return(T);
    else return(F);
}

pointer PR_LISTP() /* list? */
{
    if (list_length(args[1]) < 0) return (F);
    else return (T);
}

pointer PR_EQ() /* eq? */
{    
    if (args[1] == args[2]) return(T);
    else return(F);
}

pointer PR_EQV() /* eqv? */
{    
    if (eqv(args[1], args[2])) return(T);
    else return(F);
}

pointer PR_EQUAL() /* equal? */
{    
    if (equal(args[1], args[2])) return(T);
    else return(F);
}

pointer PR_WRITE() /* write */
{
    pointer x;
    
    if (nargs >= 3) {
        if (!isoutput(x = args[2]))
            Error("write -- second arg is not output port");
    } else x = cur_output;
    if (outputstatus(x) == 0)
       Error("write -- output port is not active");  
    printlist(args[1], x, 1, 1);
    return(T);
}

pointer PR_DISPLAY() /* display */
{    
    pointer x;
    
    if (nargs >= 3) {
        if (!isoutput(x = args[2]))
            Error("display -- second arg is not output port");
    } else x = cur_output;
    if (outputstatus(x) == 0)
       Error("display -- output port is not active");  
    printlist(args[1], x, 0, 1);
    return(T);
}

pointer PR_NEWLINE() /* newline */
{    
    pointer x;
    
    if (nargs >= 2) {
        if (!isoutput(x = args[1]))
            Error("newline -- second arg is not output port");
    } else x = cur_output;
    if (outputstatus(x) == 0)
       Error("newline -- output port is not active");
#ifndef THINK_C_DA
    PrintChar(x, '\n'); 
#else
    PrintChar(x, '\r');
#endif
    return(T);
}

pointer PR_WRCHAR() /* write-char */
{    
    pointer x;
    
   if (nargs >= 3) {
        if (!isoutput(x = args[2]))
            Error("write-char -- second arg is not output port");
    } else x = cur_output;
    if (outputstatus(x) == 0)
       Error("write-char -- output port is not active");
    if (!ischar(args[1]))
       Error("write-char -- first arg is not character");
    PrintChar(x, (char)cvalue(args[1])); 
    return(T);
}

pointer PR_RESET() /* reset */
{
    reset();
    return (T);
}

pointer PR_LENGTH() /* length */
{
    long v = list_length(args[1]);
    
    if (v < 0) Error("length -- arg is not list");
    return(mki_f_number(v));
}
                
pointer PR_ASSQ() /* assq */
{
    register pointer x = args[1], y = args[2];
    
    for (; ispair(y); y = cdr(y)) {
        if (!ispair(car(y)))
            Error("assq -- unable to handle non pair element");
        if (x == caar(y)) break;
    }
    
    if (ispair(y)) return(car(y));
    else return(F);
}

pointer PR_ASSV() /* assv */
{
    register pointer x = args[1], y = args[2];

    for (; ispair(y); y = cdr(y)) {
        if (!ispair(car(y)))
            Error("assv -- unable to handle non pair element");
        if (eqv(x, caar(y))) break;
    }
    if (ispair(y)) return(car(y));
    else return(F);
}

pointer PR_ASSOC() /* assoc */
{
    register pointer x = args[1], y = args[2];

    for (; ispair(y); y = cdr(y)) {
        if (!ispair(car(y)))
            Error("assoc -- unable to handle non pair element");
        if (equal(x, caar(y))) break;
    }
    
    if (ispair(y)) return(car(y));
    else return(F);
}

pointer PR_MEMQ() /* memq */
{
    register pointer x = args[1], y = args[2];

    for (; ispair(y); y = cdr(y))
        if (x == car(y)) break;
    
    if (ispair(y)) return(y);
    else return(F);
}

pointer PR_MEMV() /* memv */
{
    register pointer x = args[1], y = args[2];

    for (; ispair(y); y = cdr(y))
        if (eqv(x, car(y))) break;
    
    if (ispair(y)) return(y);
    else return(F);
}

pointer PR_MEMBER() /* member */
{
    register pointer x = args[1], y = args[2];

    for (; ispair(y); y = cdr(y))
        if (equal(x, car(y))) break;
    
    if (ispair(y)) return(y);
    else return(F);
}

/* debugged by Akira Kida */
pointer PR_LISTWIDTH() /* list-width */
{ return(mki_f_number(printlist(args[1], NULL, 1, 0))); }

pointer PR_PUT() /* put */
{
    register pointer x, y, z;

    if (!hasprop(args[1]) || !hasprop(args[2]))
        Error("put -- Illegal args");
    for (x = symprop(args[1]), y = args[2] ; x != NIL ; x = cdr(x))
        if (caar(x) == y) break;
    if (x != NIL) cdar(x) = args[3];
    else {
        z = cons(y, args[3]);
        symprop(args[1]) = cons(z, symprop(args[1]));
    }
    return(T);
}

pointer PR_GET() /* get */
{
    register pointer x, y;

    if (!hasprop(args[1]) || !hasprop(args[2]))
        Error("get -- Illegal args");
    for (x = symprop(args[1]), y = args[2] ; x != NIL ; x = cdr(x))
        if (caar(x) == y) break;
    if (x != NIL) return(cdar(x));
    else return(F);
}

pointer PR_CHEQ()  /* char=? */
{    
    if (!ischar(args[1]) || !ischar(args[2]))
        Error("char=? -- Illegal args");
    if (cvalue(args[1]) == cvalue(args[2])) return(T);
    else return(F);
}
   
pointer PR_CHLESS() /* char<? */
{    
    if (!ischar(args[1]) || !ischar(args[2]))
        Error("char<? -- Illegal args");
    if (cvalue(args[1]) < cvalue(args[2])) return(T);
    else return(F);
}

pointer PR_CHGRE()  /* char>? */
{    
    if (!ischar(args[1]) || !ischar(args[2]))
        Error("char>? -- Illegal args");
    if (cvalue(args[1]) > cvalue(args[2])) return(T);
    else return(F);
}

pointer PR_CHLEQ()  /* char<=? */
{    
    if (!ischar(args[1]) || !ischar(args[2]))
        Error("char<=? -- Illegal args");
    if (cvalue(args[1]) <= cvalue(args[2])) return(T);
    else return(F);
}

pointer PR_CHGEQ()  /* char>=? */
{    
    if (!ischar(args[1]) || !ischar(args[2]))
        Error("char>=? -- Illegal args");
    if (cvalue(args[1]) >= cvalue(args[2])) return(T);
    else return(F);
}

pointer PR_CCIEQ()  /* char-ci=? */
{    
    if (!ischar(args[1]) || !ischar(args[2]))
        Error("char-ci=? -- Illegal args");
    if (LowerCase((int)cvalue(args[1])) == LowerCase((int)cvalue(args[2])))
        return(T);
    else return(F);
}
   
pointer PR_CCILS() /* char-ci<? */
{    
    if (!ischar(args[1]) || !ischar(args[2]))
        Error("char-ci<? -- Illegal args");
    if (LowerCase((int)cvalue(args[1])) < LowerCase((int)cvalue(args[2])))
        return(T);
    else return(F);
}

pointer PR_CCIGRE()  /* char-ci>? */
{    
    if (!ischar(args[1]) || !ischar(args[2]))
        Error("char-ci>? -- Illegal args");
    if (LowerCase((int)cvalue(args[1])) > LowerCase((int)cvalue(args[2])))
        return(T);
    else return(F);
}

pointer PR_CCILEQ()  /* char-ci<=? */
{    
    if (!ischar(args[1]) || !ischar(args[2]))
        Error("char-ci<=? -- Illegal args");
    if (LowerCase((int)cvalue(args[1])) <= LowerCase((int)cvalue(args[2])))
        return(T);
    else return(F);
}

pointer PR_CCIGEQ()  /* char-ci>=? */
{    
    if (!ischar(args[1]) || !ischar(args[2]))
        Error("char-ci>=? -- Illegal args");
    if (LowerCase((int)cvalue(args[1])) >= LowerCase((int)cvalue(args[2])))
        return(T);
    else return(F);
}

pointer PR_CHALPHA() /* char-alphabetic? */
{
    int ch;
    if (!ischar(args[1]))
        Error("char-alphabetic? -- arg is not char");
    ch = cvalue(args[1]);
    if (('a' <= ch && ch <= 'z') || ('A' <= ch && ch <= 'Z'))
        return (T);
    else
        return (F);
}

pointer PR_CHNUMP() /* char-numeric? */
{
    int ch;
    if (!ischar(args[1]))
        Error("char-numeric? -- arg is not char");
    ch = cvalue(args[1]);
    if ('0' <= ch && ch <= '9')
        return (T);
    else
        return (F);
}

pointer PR_CHWHITE() /* char-whitespace? */
{
    int ch;
    if (!ischar(args[1]))
        Error("char-whitespace? -- arg is not char");
    ch = cvalue(args[1]);
    if (isspace(ch))
        return (T);
    else
        return (F);
}

pointer PR_CHUPPER() /* char-upper-case? */
{
    int ch;
    if (!ischar(args[1]))
        Error("char-upper-case? -- arg is not char");
    ch = cvalue(args[1]);
    if ('A' <= ch && ch <= 'Z')
        return (T);
    else
        return (F);
}

pointer PR_CHLOWER() /* char-lower-case? */
{
    int ch;
    if (!ischar(args[1]))
        Error("char-lower-case? -- arg is not char");
    ch = cvalue(args[1]);
    if ('a' <= ch && ch <= 'z')
        return (T);
    else
        return (F);
}

pointer PR_CH2INT() /* char->integer */
{
    if (!ischar(args[1]))
        Error("char->integer -- arg is not char");
    return (mki_f_number((long)cvalue(args[1])));
}

pointer PR_INT2CHAR() /* integer->char */
{   
    long v;
    
    if (!isinteger(args[1]))
        Error("integer->char -- arg is not integer");
    v = (get_longint_from_integer(args[1]))%CHARMAX;
    return((pointer)&char_table[v]);
}
    
pointer PR_CHUP() /* char-upcase */
{    
    return((pointer)&char_table[UpperCase((int)cvalue(args[1]))]);
}
    
pointer PR_CHDOWN() /* char-downcase */
{    
    return((pointer)&char_table[LowerCase((int)cvalue(args[1]))]);
}

pointer PR_MKSTR() /* make-string */
{
    return(mk_string(NULL, get_longint_from_integer(args[1]), 
                     (nargs >= 3) ? cvalue(args[2]) : ' '));
}   

pointer PR_SLEN() /* string-length */
{ return (mki_f_number(strlength(args[1]))); }

pointer PR_STRREF() /* string-ref */
{
    char *str = strvalue(args[1]);
    long v = get_longint_from_integer(args[2]);
    if (v >= strlength(args[1])) 
        Error("string-ref -- index is out of range");
    return ((pointer)&char_table[str[v]]);
}

pointer PR_STRSET() /* string-set! */
{
    char *str = strvalue(args[1]);
    long v = get_longint_from_integer(args[2]);
    if (v >= strlength(args[1]))
        Error("string-set! -- index is out of range");    
    str[v] = cvalue(args[3]);
    return (T);
}

pointer PR_STREQ() /* string=? */
{    
    if (!isstring(args[1]) || !isstring(args[2])) return (F);
    else if (str_eqv(strvalue(args[1]), strlength(args[1]),
                        strvalue(args[2]), strlength(args[2])))
        return (T);
    else return (F);
}

pointer PR_SLESS() /* string<? */
{    
    if (!isstring(args[1]) || !isstring(args[2])) return (F);
    else if (str_cmp(strvalue(args[1]), strlength(args[1]),
                            strvalue(args[2]), strlength(args[2])) < 0)
        return (T);
    else return (F);
}

pointer PR_STRGRE()  /* string>? */
{
    if (!isstring(args[1]) || !isstring(args[2])) return (F);
    else if (str_cmp(strvalue(args[1]), strlength(args[1]),
                            strvalue(args[2]), strlength(args[2])) > 0)
        return (T);
    else return (F);
}

pointer PR_SLEQ()  /* string<=? */
{    
    if (!isstring(args[1]) || !isstring(args[2])) return (F);
    else if (str_cmp(strvalue(args[1]), strlength(args[1]),
                            strvalue(args[2]), strlength(args[2])) <= 0)
        return (T);
    else return (F);
}

pointer PR_STRGEQ()  /* string>=? */
{    
    if (!isstring(args[1]) || !isstring(args[2])) return (F);
    else if (str_cmp(strvalue(args[1]), strlength(args[1]),
                            strvalue(args[2]), strlength(args[2])) >= 0)
        return (T);
    else return (F);
}

pointer PR_SCIEQ() /* string-ci=? */
{    
    if (!isstring(args[1]) || !isstring(args[2])) return (F);
    else if (str_cmp_ci(strvalue(args[1]), strlength(args[1]),
                               strvalue(args[2]), strlength(args[2])) == 0)
        return (T);
    else return (F);
}

pointer PR_SCILS() /* string-ci<? */
{    
    if (!isstring(args[1]) || !isstring(args[2])) return (F);
    else if (str_cmp_ci(strvalue(args[1]), strlength(args[1]),
                               strvalue(args[2]), strlength(args[2])) < 0)
        return (T);
    else return (F);
}

pointer PR_SCIGRE()  /* string-ci>? */
{
    if (!isstring(args[1]) || !isstring(args[2])) return (F);
    else if (str_cmp_ci(strvalue(args[1]), strlength(args[1]),
                               strvalue(args[2]), strlength(args[2])) > 0)
        return (T);
    else return (F);
}

pointer PR_SCILEQ()  /* string-ci<=? */
{    
    if (!isstring(args[1]) || !isstring(args[2])) return (F);
    else if (str_cmp_ci(strvalue(args[1]), strlength(args[1]),
                               strvalue(args[2]), strlength(args[2])) <= 0)
        return (T);
    else return (F);
}

pointer PR_SCIGEQ()  /* string-ci>=? */
{    
    if (!isstring(args[1]) || !isstring(args[2])) return (F);
    else if (str_cmp_ci(strvalue(args[1]), strlength(args[1]),
                               strvalue(args[2]), strlength(args[2])) >= 0)
        return (T);
    else return (F);
}

pointer PR_SUBSTR() /* substring */
{
    long i = get_longint_from_integer(args[2]);
    long v = get_longint_from_integer(args[3]);
    if (i < 0 || i > v || v > strlength(args[1]))
        Error("substring -- Illegal index");
    return (mk_string(&(strvalue(args[1]))[i], v - i, '\0'));
}    

pointer PR_STRAPP() /* string-append */
{
    pointer x;
    char *new, *old;
    register long i, j, v;
    
    for (j = 1, v = 0 ; j < nargs ; j++) {
        if (!isstring(args[j]))
            Error("string-append -- Illegal arg");
        v += strlength(args[j]);
    }
    x = mk_string(NULL, v, '\0');
    new = strvalue(x);
    for (j = 1; j < nargs ; j++) {
        for (i = 0, v = strlength(args[j]), old = strvalue(args[j]) ;
             i < v ; i++, new++)
            *new = old[i];
    }
    
    return(x);
}

pointer PR_STR2NUM() /* string->number */
{
    char *numstr;
    pointer x;
    uShort b;
    
    if (!isstring(args[1])) Error("string->number -- Illegal arg");
    if (nargs >= 3)
        b = (uShort)get_longint_from_integer(args[2]);
    else b = 10;
    if (b != 2 && b != 8 && b != 10 && b != 16)
       Error("string->number -- Illegal radix");
    numstr = strvalue(args[1]);
    switch (numtype(numstr, b)) {
    case NUM_TYPE_INT:
        x = str_to_integer(args[1], b);
        break;
    case NUM_TYPE_REAL:
        x = mk_real(atof(numstr));
        break;
    default:
        x = F;
    }
    
    return(x);
}

pointer PR_NUM2STR() /* number->string */
{
    pointer x;
    uShort b;
    
    if (isinteger(args[1])) {
        if (nargs >= 3)
            b = (uShort)get_longint_from_integer(args[2]);
        else b = 10;
        if (b != 2 && b != 8 && b != 10 && b != 16)
            Error("number->string -- Illegal radix");
        x = int_to_string(args[1], b);
    } else if (isreal(args[1])) {
        sprintf(&strbuff[0], "%g", rvalue(args[1]));
        x = mk_string(&strbuff[0], (long)strlen(&strbuff[0]), '\0');
    } else
        Error("number->string -- Illegal arg");
    
    return(x);
}

pointer PR_SYM2STR() /* symbol->string */
{
    if (!issymbol(args[1]))
        Error("symbol->string -- Illegal arg");
    return (mk_string(symname(args[1]), symlen(args[1]), '\0'));
}   

pointer PR_STR2SYM() /* string->symbol */
{
    if (!isstring(args[1]))
        Error("string->symbol -- Illegal arg");
    return (mk_sym(strvalue(args[1]), strlength(args[1]), 0));
}

pointer PR_VECTORP() /* vector? */
{    
    if (isvector(args[1])) return(T);
    else return(F);
}

pointer PR_MKVECTOR() /* make-vector */
{
    return (mk_vector(get_longint_from_integer(args[1]), 
                     (nargs > 2) ? args[2] : NIL));
}

pointer PR_VECLEN() /* vector-length */
{ return(mki_f_number(vector_len(args[1]))); }

pointer PR_VECREF() /* vector-ref */
{
    pointer *vec = vector_vec(args[1]);
    long v = get_longint_from_integer(args[2]);
    if (v >= vector_len(args[1]))
        Error("vector-ref -- index is out of range");
    return (vec[v]);
}

pointer PR_VECSET() /* vector-set! */
{
    pointer *vec = vector_vec(args[1]);
    long v = get_longint_from_integer(args[2]);
    if (v >= vector_len(args[1]))
        Error("vector-set! -- index is out of range");
    vec[v] = args[3];
    return (T);
}

pointer PR_VECLIST() /* vector->list */
{
    pointer x;
    pointer *y;	/* by A.Kida, to get rid of the BUG of MS-C6 */
    long v;
    
    if (!isvector(args[1]))
        Error("vector->list -- arg is not vector");
    y = vector_vec(args[1]);
    for (x = NIL, v = vector_len(args[1])-1 ; v >= 0 ; v--)
        x = cons(y[v], x);
    
    return (x);
}

pointer PR_LISTVEC() /* list->vector */
{
    pointer x, y, *vec;
    long v, i;
    
    if ((v = list_length(args[1])) < 0)
        Error("list->vector -- arg is not list");
    x = mk_vector(v, NIL);
    vec = vector_vec(x);
    for (i = 0, y = args[1] ; i < v ; y = cdr(y), i++)
        vec[i] = car(y);
    
    return (x);
}

pointer PR_EOFP() /* eof-object? */
{
    if (args[1] == eof_object) return (T);
    else return (F);
}

pointer PR_INPUTP() /* input-port? */
{    
    if (isinput(args[1])) return(T);
    else return(F);
}

pointer PR_CURINPUT() /* current-input-port */
{ return(cur_input); }

pointer PR_OPENINPUT() /* open-input-file */
#ifndef THINK_C_DA
{
    FILE *fp;
    
    if (!isstring(args[1]))
        Error("open-input-file -- arg is not string");
    if (input_ports >= MAXFILE)
        Error("open-input-file -- too many input-port");
    fp = fopen(strvalue(args[1]), "r");
    if (fp == NULL)
        Error("open-input-file -- unable to open %s", strvalue(args[1]));
    return (mk_input(fp, 0));
}
#else
{
    int err, ref;
    char *name;
    
    if (!isstring(args[1]))
        Error("open-input-file -- arg is not string");
    if (input_ports >= MAXFILE)
        Error("open-input-file -- too many input-port");
    name = strvalue(args[1]);
    CtoPstr(name);
    err = OpenReadFile(name, &ref);
    PtoCstr(name);
    if (err != noErr) {
        Error("open-input-file -- unable to open %s", name);
    }
    return (mk_input(ref, 0));
}
#endif

pointer PR_INPCLOSE() /* close-input-port */
{
    long i;
    
    if (!isinput(args[1]))
        Error("close-input-port -- arg is not input-port");
    if (inputstatus(args[1]) == 0) return (T);
    i = inputindex(args[1]);
    if (i < 0 || i >= input_ports)
        Error("close-input-port -- protected input-port");
#ifndef THINK_C_DA
    fclose(inputfile(args[1]));
#else
    FSClose(inputfile(args[1]));
#endif
    inputstatus(args[1]) = 0;
    for (; i < input_ports ; i++)
        i_port_list[i] = i_port_list[i+1];
    --input_ports;
    
    return(T);
}

pointer PR_OUTPUTP() /* output-port? */
{    
    if (isoutput(args[1])) return(T);
    else return(F);
}

pointer PR_CUROUTPUT() /* current-output-port */
{ return(cur_output); }

pointer PR_OPENOUTPUT() /* open-output-file */
#ifndef THINK_C_DA
{
    FILE *fp;
    
    if (!isstring(args[1]))
        Error("open-output-file -- arg is not string");
    if (output_ports >= MAXFILE)
        Error("open-output-file -- too many output-port");
    fp = fopen(strvalue(args[1]), "w");
    if (fp == NULL)
        Error("open-output-file -- unable to open %s", strvalue(args[1]));
    return (mk_output(fp));
}
#else
{
    int err, ref;
    char *name;
    
    if (!isstring(args[1]))
        Error("open-output-file -- arg is not string");
    if (output_ports >= MAXFILE)
        Error("open-output-file -- too many output-port");
    name = strvalue(args[1]);
    CtoPstr(name);
    err = OpenWriteFile(name, &ref);
    PtoCstr(name);
    if (err != noErr) {
        Error("open-output-file -- unable to open %s", name);
    }
    return (mk_output(ref));
}
#endif

pointer PR_OUTCLOSE() /* close-output-port */
{
    long i;
    
    if (!isoutput(args[1]))
        Error("close-output-port -- arg is not output-port");
    if (outputstatus(args[1]) == 0) return (T);
    i = outputindex(args[1]);
    if (i < 0 || i >= output_ports)
        Error("close-output-port -- protected output-port");
#ifndef THINK_C_DA
    fclose(outputfile(args[1]));
#else
    FSClose(outputfile(args[1]));
#endif
    outputstatus(args[1]) = 0;
    for (; i < output_ports ; i++)
        o_port_list[i] = o_port_list[i+1];
    --output_ports;
    
    return(T);
}

pointer PR_EXACT() /* exact? */
{
    if (isinteger(args[1])) return (T);
    else return (F);
}

pointer PR_INEXACT() /* inexact? */
{
    if (isreal(args[1])) return (T);
    else return (F);
}

pointer PR_GENSYM() /* gensym */
{
    if (nargs >= 2) {
        if (!isstring(args[1])) 
            Error("gensym -- first argument must be string");
        return (gensym(strvalue(args[1])));
    } else return(gensym(NULL));
}

pointer PR_GC() /* gc */
{
    gc(NIL, NIL);
    return(T);
}

pointer PR_GCVERBOSE() /* gc-verbose */
{
    if (istrue(args[1]))
        gc_verbose = 1;
    else
        gc_verbose = 0;
    
    return(T);
}

pointer PR_GETCLOSURE() /* get-closure-code */
{
    if (!isclosure(args[1]) && !ismacro(args[1]))
        Error("get-closure-code -- arg is not closure or macro");
    return (cons(LAMBDA, closure_code(args[1])));
}

pointer PR_IORESET() /* rset-io */
{
    reset_io();
    return (T);
}

pointer PR_MKBOX() /* box */
{  return (mk_box(args[1])); }

pointer PR_BOXP() /* box? */
{  return (isbox(args[1]) ? T : F); }

pointer PR_UNBOX() /* unbox */
{
    if (!isbox(args[1]))
        Error("unbox -- argument must be box", NIL);
    return (boxcontext(args[1]));
}

pointer PR_SETBOX() /* set-box! */
{
    if (!isbox(args[1]))
        Error("set-box! -- first argument must be box", NIL);
    boxcontext(args[1]) = args[2];
    return (args[2]);
}

pointer PR_EXIT() /* exit */		/*  Added by Akira Kida */
{
    longjmp(reset_jmp, 2);
}
