/*
 * print.c
 *
 * This file is part of AM-Scheme.
 * AM-Scheme is coded by Atsushi Moriwaki
 *
 * This software is completely free to copy and distribute modifed version.
 * But I would appreciate it if you left my name on the code as the author.
 *
 */
 
#include "amscm.h"

#define p_goto(a) { pr_operator = (a); goto LOOP; }

#define p_save(a, b) {                                 \
    if (stack + 2 >= stacktop) Error("stack overflow");\
    *++stack = (b);                                    \
    *++stack = (pointer)(a); }

#define p_return {                             \
    pr_operator = (int)*stack--;               \
    list = *stack--;                           \
    goto LOOP; }

/* print integer */
long printinteger(o, file, pr)
pointer o;
pointer file;
int pr;
{
    long size = integersize(o), length;
    uShort *array = integerarray(o);

    sprintf(&strbuff[0], "%d", integersign(o) * array[--size]);
    length = strlen(&strbuff[0]);
    if (pr) PrintString(file, &strbuff[0]);
    while (size) {
        sprintf(&strbuff[0], "%04d", array[--size]);
        length += strlen(&strbuff[0]);
        if (pr) PrintString(file, &strbuff[0]);
    }
    return (length);
}

/* print atoms */
long printatom(l, file, f, pr)
pointer l;
pointer file;
int f;	/* 1 if write, 0 if display */
int pr;
{
    char c, *p;
    long i, len, length;

    if (l == NIL) {
        length = 2;
        if (pr) PrintString(file, "()");
    } else if (l == T) {
        length = 2;
        if (pr) PrintString(file, "#t");
    } else if (l == F) {
         length = 2;
         if (pr) PrintString(file, "#f");
    } else if (l == eof_object) {
        length = 13;
        if (pr) PrintString(file, "#<eof-object>");
    } else if (isinteger(l)) {
        length = printinteger(l, file, pr);
    } else if (isreal(l)) {
        sprintf(&strbuff[0], "%g", rvalue(l));
        length = strlen(&strbuff[0]);
        if (pr) PrintString(file, &strbuff[0]);
    } else if (isstring(l)) {
        if (f) {   /* write */
            len = strlength(l); p = strvalue(l);
            length = len + 2;
            if (pr) PrintChar(file, '"');
            for (i = 0; i < len; i++) {
                c = p[i];
                if (!isprint(c) && c != '\n') {
                    if (pr) PrintChar(file, ' ');
                } else {	/* debugged by Akira Kida */
                    switch (c) {
                    case '\n':
                        c = 'n';  /* NOT BREAK */
                    case '\\':
                    case '"':
                        if (pr) PrintChar(file, '\\');
                        length += 1;  /* NOT BREAK */
                    default:
                        if (pr) PrintChar(file, c);
                        break;
                    }
                }
            }
            if (pr) PrintChar(file, '"');
        } else {
            length = strlength(l);
            p = strvalue(l);
            if (pr) {
                for (i = 0; i < length ; i++) {
                        c = p[i];
                        if (isprint(c) || c == '\n')
                             PrintChar(file, c);
                        else
                             PrintChar(file, ' ');
                }
            }
        }
    } else if (issymbol(l)) {
        length = symlen(l);
        if (pr) PrStrWithLen(file, symname(l), symlen(l));
    } else if (ischar(l)) {
        if (f) {
            if (isprint(c = cvalue(l))) sprintf(&strbuff[0], "#\\%c", c);
            else sprintf(&strbuff[0], OCTALFORMAT, (int)c);
            length = strlen(&strbuff[0]);
            if (pr) PrintString(file, &strbuff[0]);
        } else {
            sprintf(&strbuff[0], "%c", cvalue(l));
            length = strlen(&strbuff[0]);
            if (pr) PrintString(file, &strbuff[0]);
        }
    } else if (isprim(l)) {
        sprintf(&strbuff[0], "#<PRIMITIVE %s>", primname(l));
        length = strlen(&strbuff[0]);
        if (pr) PrintString(file, &strbuff[0]);
    } else if (ismacro(l)) {
        length = 8;
        if (pr) PrintString(file, "#<MACRO>");
    } else if (isclosure(l)) {
        length = 10;
        if (pr) PrintString(file, "#<CLOSURE>");
    } else if (iscontinuation(l)) {
        length = 15;
        if (pr) PrintString(file, "#<CONTINUATION>");
    } else if (isinput(l)) {
        length = 13;
        if (pr) PrintString(file, "#<INPUT PORT>");
    } else if (isoutput(l)) {
        length = 14;
        if (pr) PrintString(file, "#<OUTPUT PORT>");
    }
    return (length);
}

/* print list cell */
long printlist(list, file, f, pr)
pointer list;
pointer file;
int f, pr;
{
    register pr_operator = OP_PRBEGIN;
    register long length = 0;
    
LOOP:
    switch(pr_operator) {
 
case OP_PRBEGIN:
    g_push(&list) g_push(&file)
    p_save(OP_PREND, NIL)
    p_goto(OP_PRLST)

case OP_PREND:
    g_pop(2)
    return (length);
    
case OP_PRLST:
    if (!ispair(list)) {
        if (isvector(list)) {
            long i; pointer *vec;
            length += 2;
            if (pr) PrintString(file, "#(");
            if ((i = vector_len(list)) == 0) {
                length +=1;
                if (pr) PrintChar(file, ')');
                p_return
            }
            --i; s_push(NULL)
            for (i = vector_len(list) - 1, vec = vector_vec(list) ;
                 i >= 1 ; i--)
                s_push(vec[i])
            p_save(OP_PRVEC, NIL)
            list = vec[0];
            p_goto(OP_PRLST)
        } else if (isbox(list)) {
            length += 6;
            if (pr) PrintString(file, "#<BOX ");
            p_save(OP_PRBOX, NIL)
            list = boxcontext(list);
            p_goto(OP_PRLST)
        } else {
            length += printatom(list, file, f, pr);
            p_return
        }
    } else {
        if (pr) PrintChar(file, '(');
        length += 1;
        p_save(OP_PRCDR, cdr(list))
        list = car(list);
        p_goto(OP_PRLST)
    }

case OP_PRCDR:
    if (ispair(list)) {
        p_save(OP_PRCDR, cdr(list))
        length += 1;
        if (pr) PrintChar(file, ' ');
        list = car(list);
        p_goto(OP_PRLST)
    } else {
        if (list != NIL) {
            length += 3;
            if (pr) PrintString(file, " . ");
            p_save(OP_PRCDR, NIL)
            p_goto(OP_PRLST)
        }
        length += 1;
        if (pr) PrintChar(file, ')');
        p_return
    }

case OP_PRVEC:
    if ((list = s_pop()) == NULL) {
        length += 1;
        if (pr) PrintChar(file, ')');
        p_return
    } else {
        p_save(OP_PRVEC, NIL)
        length += 1;
        if (pr) PrintChar(file, ' ');
        p_goto(OP_PRLST)
    }

case OP_PRBOX:
    list += 1;
    if (pr) PrintChar(file, '>');
    p_return
    
default:
    Error("Illegal print operator %d", pr_operator);
    }
}
    
