/*
 * read.c
 *
 * This file is part of AM-Scheme.
 * AM-Scheme is coded by Atsushi Moriwaki
 *
 * This software is completely free to copy and distribute modifed version.
 * But I would appreciate it if you left my name on the code as the author.
 *
 */
 
#include "amscm.h"

pointer r_value;
int Is_read_complete;

read_op_type *read_stack;
read_op_type *btm_rd_stack;
read_op_type *top_rd_stack;

read_char(l)
pointer l;
{
    int ch;
    if (inputpos(l) >= inputeol(l)) {
#ifndef THINK_C_DA
        if (Get_Line(l) == EOF) {
            ch = EOF;
            goto END;
        }
    }
#else
        if (l == std_input) {
            ch = '\0';
            goto END;
        } else if (Get_Line(l) == EOF) {
            ch = EOF;
            goto END;
        }
    }
#endif
#ifndef THINK_C_DA
     ch = inputbuffer(l)[inputpos(l)];
#else
     if (inputotherbuff(l) != NULL)
         ch = inputotherbuff(l)[inputpos(l)];
     else
         ch = inputbuffer(l)[inputpos(l)];
#endif
END:
    inputpos(l) += 1;
    return (ch);
}

back_char(l)
pointer l;
{
    if (inputpos(l) > 0) inputpos(l) -= 1;
}

read_white(l)
pointer l;
{
    char ch;
    
    while (inputpos(l) < inputeol(l)) {
        ch = read_char(l);
        if (!isspace(ch)) {
            back_char(l);
            break;
        }
    }
}

/* check c is delimiter */
isdelim(s, c)
char *s;
int c;
{
    if (c == '\0') return (1);
    while (*s) if (*s++ == c) return (1);
    return (0);
}

/* read chacters to delimiter */
readstr(delim, l)
char *delim;
pointer l;
{
    char *p = strbuff;
    int ch;

    for (;;) {
        if ((ch = read_char(l)) == EOF) {
            *p = '\0';
            break;
        } else if (ch == '\0') {
            *p = '\0';
            break;
        } else if (isdelim(delim, ch)) {
            back_char(l);
            *p = '\0';
            break;
        } else {
            *p++ = ch;
        }
    }
}

/* read string expression "xxx...xxx" */ /* changed by Akira Kida */
readstrexp(l)
pointer l;
{
    char *p = strbuff;
    int ch;

    while ((ch = read_char(l)) != '"') {
        if (ch == '\0' || ch == EOF)
            Error("reader -- Illegal string expression");
        *p++ = ch;
        if (ch == '\\') {
            switch ((ch = read_char(l))) {
            case 'n':
                ch = '\n';
                break;
            case '\\':
            case '"':
                /* do nothing */
                break;
            default:
                if (!isprint(ch))
                    ch = ' ';
                ++p;
                break;
            }
            p[-1] = ch;
        }
    }
    *p = '\0';
}

/* read constant expression */
readconstexp(delim, l)
char *delim;
pointer l;
{
    char *p = strbuff;
    int ch = read_char(l), count = 0;

    for ( ; !isdelim(delim, ch) && ch != '\0' && ch != EOF ; ) {
        *p++ = ch; ++count;
        ch = read_char(l);
    }
    back_char(l);
    if (count == 2 && strbuff[1] == '\\') {
        ch = read_char(l);
        if (ch == ' ' || ch == ';' || ch == '(' || ch == ')') *p++ = ch;
        else back_char(l);
    }
    *p = '\0';
}    

/* get token */
tok_type get_token(l)
pointer l;
{
    int ch;
AGAIN:
    if ((ch = read_char(l)) == EOF) return (TOK_EOF);
    if (isspace(ch)) goto AGAIN;
    switch (ch) {
        case '('  : return (TOK_LPAREN);
        case ')'  : return (TOK_RPAREN);
        case '.'  : if (isspace(read_char(l))) return (TOK_DOT);
                    else {
                        back_char(l); back_char(l);
                        readstr("();\t\n\r ", l);
                        return (TOK_ATOM);
                    }
        case '\'' : return (TOK_QUOTE);
        case '`'  : return (TOK_BQUOTE);
        case ','  : if (read_char(l) == '@') return (TOK_ATMARK);
                    else { back_char(l); return (TOK_COMMA); }
        case '"'  : readstrexp(l);
                    return (TOK_DQUOTE);
        case '#'  : if (read_char(l) == '(')
                        return (TOK_VECTOR);
                    else {
                        back_char(l); back_char(l);
                        readconstexp("();\t\n\r ", l);
                        return (TOK_SHARP);
                    }
        case ';'  : for (;;) {
                        ch = read_char(l);
                        if (ch == EOF) return (TOK_EOF);
                        else if (ch == '\n' || ch == '\r' || ch == '\0')
                            break;
                    }
                    goto AGAIN;
        case '\0' : return (TOK_EOL);
        default   : back_char(l);
                    readstr("();\t\n\r ", l);
                    return (TOK_ATOM);
    }
}

/* get list form stack */
pointer get_list_from_stack(ini)
pointer ini;
{
    register pointer p;
    
    for (p = ini; *stack != NULL ; --stack)
        p = cons(*stack, p);
    --stack;
    
    return (p);
}

/* get vector form stack */
pointer get_vector_from_stack()
{
    long i, len;
    register pointer x, *b, *vec;
    
    for ( b = stack, len = 0; *b != NULL ; --b, ++len)
        ;
    ++b;
    x = mk_vector(len, NIL);
    vec = vector_vec(x);
    for (i = 0 ; i < len ; i++)
        vec[i] = b[i];
    stack = b - 2;
    return (x);
}

/* read S-expression */
Read_Cycle(tok)
tok_type tok;
{
    switch (tok) {
    case TOK_LPAREN:
        s_push(NULL)
        r_push(OP_RDLST)
        return;
    case TOK_RPAREN:
        switch (r_pop()) {
        case OP_RDLST:
            r_value = get_list_from_stack(NIL);
            break;
        case OP_RDDOT:
            if (r_pop() != OP_RDLST)
                Error("reader -- Illegal dot expression");
            r_value = get_list_from_stack(r_value);
            break;
        case OP_RDVEC:
            r_value = get_vector_from_stack();
            break;
        default:
            Error("reader -- Illegal ')'");
        }
        break;
    case TOK_DOT:
        r_push(OP_RDDOT)
        return;
    case TOK_VECTOR:
        s_push(NULL)
        r_push(OP_RDVEC)
        return;
    case TOK_QUOTE: 
        r_push(OP_RDQUT)
        return;
    case TOK_BQUOTE: 
        r_push(OP_RDQQ)
        return;   
    case TOK_COMMA: 
        r_push(OP_RDUQ)
        return;    
    case TOK_ATMARK:
        r_push(OP_RDUQSP)
        return;
    case TOK_ATOM:
        r_value = mk_atom(&strbuff[0]);
        break;
    case TOK_DQUOTE:
        r_value =  mk_string(&strbuff[0], (long)strlen(&strbuff[0]), '\0');
        break;
    case TOK_SHARP:
        r_value = mk_const(&strbuff[0]);
        break;
    case TOK_EOF:
        if (r_pop() != OP_RDEND)
            Error("reader -- Illegal EOF");
        else {
            Is_read_complete = 1;
            r_value = eof_object;
            return;
        }
        break;
    default:
        Error("reader -- Illegal token %d", tok);
    }

AGAIN:    
    switch (r_pop()) {
    case OP_RDLST:
        s_push(r_value)
        r_push(OP_RDLST)
        return;
    case OP_RDDOT:
        r_push(OP_RDDOT)
        return;
    case OP_RDVEC:
        s_push(r_value)
        r_push(OP_RDVEC)
        return;    
    case OP_RDQUT:
        r_value = cons(QUOTE, cons(r_value, NIL));
        goto AGAIN;
    case OP_RDQQ:
        r_value = cons(QQUOTE, cons(r_value, NIL));
        goto AGAIN;
    case OP_RDUQ:
        r_value = cons(UNQUOTE, cons(r_value, NIL));
        goto AGAIN;
    case OP_RDUQSP:
        r_value = cons(UNQUOTESP, cons(r_value, NIL));
        goto AGAIN;
    case OP_RDEND:
        Is_read_complete = 1;
        return;
    default:
        Error("reader -- Illegal read operator");
    }
}
